/****************************************************************************/
/*                 MASS project: TURBINA module file                        */
/*                            atmos.c                                       */
/*                    Description is given in atmos.h                       */
/* Editor tabular size: 4                                                   */
/****************************************************************************/
 
/* Insert the definition ATMOSTEST in compilation for the module self-tests:
gcc -c atmos.c .... -DATMOSTEST=1 */
//#define ATMOSTEST
 
#include "atmos.h"

#ifndef SQR
#define SQR(a) ((a)*(a))
#endif

#ifdef ATMOSTEST
#define NTEST 10

/* Display the intermediate information for checking the define-settings */
/** Print names of shifted weights in use, number of base 
functions after small singulars rejection etc. */
//#define PRNBASEINFO	
/** Print the coefficients of decomposition by orthonormal basis: */
//#define PRNCOEF
/** Print the plot data for "atmos | xmgrace -pipe" for examination of
PRNAPPROX-th moment approximation: */
//#define PRNAPPROX 3

#if defined PRNAPPROX || defined PRNCOEF || defined PRNBASEINFO
	#undef NTEST
	#define NTEST 1
#endif
/*---------------------------- TEST PART ---------------------------------*/

/*----------- Comes normally from configuration file: cf_... vars ----------*/

/* Apertures: */
/* number of apertures in segmentor : */
int cf_naper = 4;
/* Physical size of segmentor annular apertures [cm] : */
double cf_aper_inn[] = { 0.000, 0.102, 0.188, 0.354};
double cf_aper_out[] = { 0.100, 0.185, 0.350, 0.650};
//double cf_aper_inn[] = { 0.000, 0.100};
//double cf_aper_out[] = { 0.100, 0.200};
/* Magnification of the system: */
double cf_magnif = 20.0 ;

/* Proportional Altitude grid: */
double cf_z0 = 0, cf_zmax = 25.0, cf_dz = 0.2, cf_dzmin = 0.2;
/* Equidistant Altitude grid: */
//double cf_z0 = 0, cf_zmax = 25.0, cf_dz = 0.5, cf_dzmin = 0.0;

/* Alternative grid: to interpolate to */
double int_z0 = 0, int_zmax = 20.0, int_dz = 0.02, int_dzmin = 0;

double cf_shift = -0.5 ;	/* Shift weight functions by cf_shift kilometers */

#define NSAMIDX 10*2+1
double samidx[NSAMIDX] = {
#include "pachon_idx.dat2"	/* uncomment one line in this file */
,0.001 /* Shifted DESI */};
#define IDXRELERR 0.01	/* Relative error of indices */
#define NDESI 1

#define FILENAMESZ 80
#define CF_NMEAS 10
char * at_fw = "weight.new" ;

char at_fcrv[] = "mass.crv" ;	/* System response curve */
//char at_fsed[] = "../weif/spectra/a05.sp" ;	/* Energy distrib. file */
char at_fsed[] = "a05.sed" ;	/* Energy distrib. file */

int main( int ac, char **av ) {

int err, i, test, idum = (-13), meth, meas ;
double idx[NSAMIDX], aidx[NSAMIDX], eidx[NSAMIDX], chi2 ;
const char foutname[] = "atmos.out" ;
FILE * fout ;
	
if ((fout = fopen(foutname,"at"))==NULL) {
	fprintf(stderr,"%s open error\n",foutname);
	return -1 ;
}

if ((err=at_init(
	cf_naper, cf_magnif, cf_aper_inn, cf_aper_out,
	cf_z0, cf_zmax, cf_dz, cf_dzmin,
	at_fcrv))!=0) {
	if (err>0)
		fprintf(stderr,"CRV error in line %d\n",err) ;
	else
		fprintf(stderr,"Init error: %s\n",sc_error(err));
	return err ;
}

for (test=0; test<NTEST; test++) {
	err = at_loadsed(at_fsed) ;
	if (err<0) {
		fprintf(stderr,"at_loadsed error: %s\n",sc_error(err));
		return err ;
	} else {
		if (err>0) {
			fprintf(stderr,"at_loadsed error: in %d-th line of %s\n",
				err,at_fsed);
			return err ;
		}
	}
	
	err = at_checkwf(at_fw) ;
	if (err) {
		printf("Start at_calcwf()\n");
		err = at_calcwf(at_fw, wf_progress) ;
		if (err) {
			fprintf(stderr,"at_calcwf() error %d: %s\n",err,sc_error(err));
			return err ;
		}
	}

	err = at_update(at_fw, "tmp.txt", 0, cf_shift) ;

	if (err) {
		fprintf(stderr,"Update error: %s\n",sc_error(err));
		return err ;
	}	
	for (i=0; i<NSAMIDX; i++) aidx[i] = eidx[i] = 0 ;
	
	if ((err=at_alloc(CF_NMEAS))<0) {
		fprintf(stderr,"at_alloc error %s\n",sc_error(err));
		return err ;
	}
	for (meas=0; meas<CF_NMEAS; meas++) {
		/* Emulate the noisy indices: */
		for (i=0; i<NSAMIDX; i++) {
			idx[i] = samidx[i]  * (1.0 + gasdev(&idum) * IDXRELERR) ; 
			aidx[i] += idx[i] ;
			chi2 = (idx[i] - samidx[i]) ;
			eidx[i] += chi2*chi2 ;
		}

		if ((err=at_calcint(idx, NULL, meas))<0) {
			fprintf(stderr,"Integral Calc error: %s\n",sc_error(err));
			return err ;
		}
	}
	for (i=0; i<NSAMIDX; i++) {
		eidx[i] = sqrt(eidx[i] / CF_NMEAS / (CF_NMEAS-1)) ;
		aidx[i] /= CF_NMEAS ;
//		printf("%d: %f: %f +- %f\n",i,samidx[i],aidx[i],eidx[i]);
	}
	for (meth=0;meth<=1;meth++) {
		if ((chi2=at_calcn2(aidx, NULL ,eidx,&(eidx[10]),meth))<0) {
			fprintf(stderr,"Cn2 Calc error: %s\n",sc_error((int)chi2));
			return err ;
		} 

		if (!meth) {
			printf("Chi2\tz_1\tz_2\tz_3\tz_4\tCn2_1\tCn2_2\tCn2_3\tCn2_4\n");			
			printf("%5.2f\t",chi2);
			for (i=0; i<at_getncn2(); i++)
			printf("%2d\t",(int)(at_getzcn2()[i]+0.5)) ;
		} else {
			printf("Chi2\tCn2_0km\tCn2_1km\tCn2_2km\tCn2_4km\tCn2_8km\tCn2_16km\n");
			printf("%5.2f\t",chi2);
		}
		for (i=0; i<at_getncn2(); i++)
			printf("%7.1e\t",at_getcn2()[i]);
		printf("\n");
	}
	if ((err=at_avgint(meas,aidx[20],eidx[20]*eidx[20],0.0))<0) {
		fprintf(stderr,"Average integrals error: %s\n",sc_error(err));
		return err ;
	}
	/* Output the parameters as a single table: */
	printf("# fSee\tSee\tTheta_0\tfHeff\tHeff\tM_2*e5\tTconst\n");
	printf("%5.2f\t%5.2f\t%5.2f\t%5.0f\t%5.0f\t",
		at_getseeing(1), at_getseeing(0), at_getisoplan(), at_getheff(1),
			at_getheff(0));
		printf("%6.3f\t%5.2f\n", at_getavgint()[AT_IISK]*1e5, at_gettimeconst());
	if ((err=at_writeint(fout,"# A"))<0) {
		fprintf(stderr,"Write atm.parameters error: %s\n",sc_error(err));
		return err ;
	}
	if ((err=at_writecn2(fout,"# C"))<0) {
		fprintf(stderr,"Write integrals error: %s\n",sc_error(err));
		return err ;
	}

	printf("#========================= %d==========================\n",test);
}
at_done() ;
	
fclose(fout) ;	
return 0 ;
}
#endif

/*------------------------ Global parameters -----------------------------*/

/** Free atmosphere seeing computed by at_calcint() */
double at_fseeing = 0, at_efseeing = 0 ;

/** Whole atmosphere seeing computed by at_calcint() */
double at_seeing = 0 , at_eseeing = 0 ;

/** Isoplanatic angle  computed by at_calcint() */
double at_isoplan = 0 , at_eisoplan = 0 ;

/** Effective altitude of the turbulence  from at_calcint(): */
double at_heff = 0 , at_eheff = 0 ;

/** Effective altitude in free atmosphere from at_calcint(): */
double at_fheff = 0 , at_efheff = 0 ;

/** atmospheric time constant [ms]  from at_calcint() */
double at_timeconst = 0 , at_etimeconst = 0 ;

/** Length of the computed Cn^2-vector from at_calcn2() */
int at_ncn2 = 0 ;

/** Cn^2-vector computed by Cn2 restoration at_calcn2() and Chi2 as last elem.*/
double * at_cn2 = NULL ;

/** Altitudes of Cn^2 layers from at_calcn2() */
double * at_zcn2 = NULL ;

/** Basic set of weight functions: */
wf_t at_wf ;

/** User-set altitude grid: */
double at_z0 = WF_Z0, at_zmax = WF_ZMAX, at_dz=WF_DZ, at_dzmin=WF_DZMIN ; 

/** Response function of the atmosphere + feeding optics + MASS system */
double * at_response = NULL ;

/* Currently loaded weight file name: */
char * at_curwffile = NULL ;

/** Response function wavelength grid */
double * at_lambda = NULL ;

/** at_lambda's and at_response' grid length: */
int at_nresponse = 0 ;

/** Current star spectrum multiplied by response curve at_response */
double * at_sed = NULL ;

/** Physical length of current star spectrum array */
int at_nsed = 0 ;

/** Two Matrices of coefficients for the moments of Cn2  */
double *** at_coef = NULL ;	/**< [0]: for non-shifted measurements, [1] 
for generalized mode measurements */

/** Number of photometric channels */
int at_nchan = 0 ;

/** Number of weight functions used in normal and generalized modes (the 
 column number of the matrix \em at_coef) */
int at_nweight[AT_NMODE] = {0,0} ;

/** Shifts by altitude by which normal (first) and generalized (both) modes 
operate */
double at_zshift[AT_NMODE] = {0,0} ;

/* Global combined matrix of weights for restoration of the profile: non-shifted
and shifted weights togather, with 1km spacing */
double ** at_wcomb = NULL ;

/* Global combined average indices and their errors for restoration of
Cn2-profile */
double * at_sicomb = NULL, * at_e2sicomb = NULL ;

/* Matrix of individual integrals */
double ** at_int = NULL ;

/* Number of allocated rows in at_int[][] */
int at_maxnmeas = -1 ;

/* Vectors of average integrals and their errors */
double * at_aint = NULL ;
double * at_e2int = NULL ;

/*============================== Functions ===============================*/

/*------------------------------ at_getseeing ----------------------------*/

double at_getseeing(int isfree) {
	if (isfree) 
		return at_fseeing ;
	else 
		return at_seeing  ;
}

/*------------------------------ at_getisoplan ---------------------------*/

double at_getisoplan(void) {
	return at_isoplan ;
}

/*------------------------------ at_gettimeconst -------------------------*/

double at_gettimeconst(void) {
	return at_timeconst ;
}

/*------------------------------ at_getisoplan ---------------------------*/

double at_getheff(int isfree) {
	if (isfree) 
		return at_fheff ;
	else 
		return at_heff ;
}

/*------------------------------ at_getncn2 -------------------------------*/

int at_getncn2(void) {
	return at_ncn2 ;
}

/*------------------------------ at_getzcn2 -------------------------------*/

double * at_getzcn2(void) {
	return at_zcn2 ;
}

/*------------------------------ at_getcn2 --------------------------------*/

double * at_getcn2(void) {
	return at_cn2 ;
}

/*------------------------------ at_getavgint ------------------------------*/

double * at_getavgint() {
	return at_aint ;
}

/*------------------------------ at_geter2int ------------------------------*/

double * at_geter2int() {
	return at_e2int ;
}

/*------------------------------ at_read2col --------------------------------*/

/**
	@brief	Read the two columns of argument and function values from file
	@param	filename	file name
	@param	col1	pointer to the array of arguments (first column data)
	@param	col2	pointer to the array of function values (second column)
	@param	nrow	pointer to the place where to put the length of allocated 
		arrays of arguments and function values
	@return	0 if Ok, <0 on general error, >0 number of line where the bad data
		are encountered
	
	This utility reads the pairs of numeric values from the text-type file
	ignoring the lines beginning with comment symbol "#". Two non-allocated
	pointers must be supplied as NULLs on start. Function counts the non-comment
	data lines, allocates the memory for arrays of argument (first column) and
	function values (second column) with lengths equal to the number of the 
	counted data lines, and then reads the columns into them.
	
	The following checks are done on each non-comment line:
	- at least two numeric tokens are given in each line (argument and function)
	- argument is a monotonously rising function
	- argument values start from some positive number
	- function is non-negative 
	- function has a positive average
	The function is thus suited for reading the spectrum-like information.
	
	Errors possible are the problem of file opening/reading or line parsing, or
	memory allocation failures. In case of error, the pointers are returned
	unallocated.
*/	
int at_read2col(const char * filename, double ** col1,
	double ** col2, int * nrow) {

#define AT_DATALINE 100       /* longest feasible line */
#define AT_NFLD 2			/* items per line in nrow file */
	FILE *f ;
	char buff[AT_DATALINE];
	char * id, *endptr;
	int nline, ndata, err;
	double prevarg, sum ;

	/* Protect from allocation on allocated pointer */
	if ((*col1!=NULL) || (*col2!=NULL)) return SC_ERNNL ;

	if (filename==NULL) return SC_ERPAR ;
	
	/* Count col1 steps: */
	if ((f = fopen(filename,"rt")) == NULL) 
		return SC_ERFIO ;
		
	ndata = nline = 0;
	while (fgets(buff,AT_DATALINE,f)!=NULL) {
		nline++ ;
		if (((id = strtok(buff, WF_FDELIM)) != NULL) && (id[0]!='#')) 
			/* it is not comment and there is at least one item */
			ndata++;
	}
	fclose(f);
	
	if (!ndata) return SC_ERNOD ;
	
	/* Reopen the file : */
	if ((f = fopen(filename,"rt")) == NULL) return SC_ERFIO ;

	/* Allocate memory and, if fails, close the file and return: */
	if ((((*col1)=dvector(0,ndata-1))==NULL) ||
		(((*col2)=dvector(0,ndata-1))==NULL) ) { 
		fclose(f); 
		return SC_ERMEM ;
	}
	*nrow = ndata ;

	err = nline = ndata = 0;
	prevarg = sum = 0 ;	/* previous argument and function \int storages */
	while (fgets(buff,AT_DATALINE,f)!=NULL) {
		nline++ ;
		if (((id = strtok(buff, WF_FDELIM)) != NULL) &&	(id[0]!='#')) {
			/* it is not comment : try to read data now: */

			/* argument: */
			(*col1)[ndata] = strtod(id, &endptr) ;
			if ((endptr==id) || ((id = strtok(NULL, WF_FDELIM)) == NULL) || 
				((*col1)[ndata]<=prevarg)) {
				/* Argument token is not a number or 
				   previous argument is not less than current or
				   function value is absent: return line number */
				err = nline ;
				break ;
			}
			
			/* function: */
			(*col2)[ndata] = strtod(id, &endptr) ;
			if ((endptr==id) || ((*col2)[ndata]<0)) {
				/* Function token is not a number or
				   function is negative */
				err = nline ;
				break ;
			}
			sum += (*col2)[ndata] ;
			prevarg = (*col1)[ndata] ;
			ndata++;
		}
	}
	fclose(f);
	/* Check that the function is positive (not zero): */
	if (!err && (sum<=0)) err = SC_ERNOD ;
	if (err) {
		/* Reading or conversion or argument/function constraints error: */
		free_dvector(*col1,0,*nrow-1);
		free_dvector(*col2,0,*nrow-1);
		return err ;
	} 
	return 0 ;
#undef AT_DATALINE
#undef AT_NFLD
}

/*-------------------------------- at_conwgrid -------------------------------*/

/**
	@brief	Convert array of optical (!) wavelengths into given scale
	@param	wgrid	array of optical wavelengths in [angstrem] or [nm]
	@param	n	its length
	@param	scale	decimal logarithm of the wavelength of the infrared end of 
		the optical range to obtain
	@return	0	or <0 on error
	
	This is an utility for conversion of the spectral curves (response function
	or spectrum) into the wavelength grid in a given units \em scale. It takes
	the value from the middle of the array and assumes that it must be in
	between 0.1 and 1mkm if it's an optical range. Then it makes the conversion by
	multiplication of the array value by an integer power of 10 which makes the
	middle array value lying in the range (0.1--1.0)*10**{scale}.
	
	\b Example
	
	Let the middle point of \em wgrid (it is \em wgrid[n/2]) be equal to 410 on
	start (i.e. the grid is in [nm]). \em Scale is given as "0", i.e. we want
	the spectrum to belong to the range 0.1--1.0, i.e. to be expressed in [mkm]
	on exit. The initial range is \em range=(int)lg(410)+1=3. So, the
	conversion coefficient will be \em 10**conv, where \em conv=scale-range=-3.
*/
int	at_conwgrid(double * wgrid, int n, int scale) {
	double conv ;
	int i ;
	if ((wgrid==NULL) || (n<1)) return SC_ERNUL ;
	
	conv = scale - ((int)log10(wgrid[n/2])+1) ;
/* printf("Power conversion = %3.0f\n",conv);*/
	conv = pow(10.0,conv);
/*printf("Conversion = %7.1e\n",conv);*/
	for (i=0; i<n; i++) wgrid[i] *= conv;
	return 0 ;
}

/*------------------------------ at_init() --------------------------------*/

int at_init(
	int nchan, double diam, double *eps_inn, double *eps_out,
	double z0, double zmax, double dz, double dzmin,
	const char * responsefile) {
#define MKM_RANGE	0	
/**< Units of optical range wavelength: 0 for [mkm], 3 for [nm], 4 for [\AA] */

	int err, nw, nz, i ;
	const char useshft[] = AT_USESHFT ;
	
	if (at_coef!=NULL) {
		/* at_done() was not called in advance! */
		at_done() ;
	}

	if ((eps_inn==NULL) || (eps_out==NULL)) 
		return SC_ERNUL ;

	/* Read the response function from the file: */
	if ((err=at_read2col(responsefile,&at_lambda,&at_response,&at_nresponse))!=0)
		return err ;
	/* Convert whatsoever (nm or angstrems) into microns: */
	if ((err=at_conwgrid(at_lambda,at_nresponse,MKM_RANGE))<0) return err ;

	/* Initialize weight structures: */
	wf_init(&at_wf) ;	

	/* Set the altitude grid and the aperture geometry in basic wf */	
	if ((err=wf_setzgrid(&at_wf, z0, zmax, dz, dzmin))<0) return err ;
	nz = wf_getnz(&at_wf) ;

	/* Assign user-supplied zgrid parameters for at_update(): */
	at_z0 = z0, at_zmax = zmax, at_dz = dz; at_dzmin = dzmin ;
	
	if ((err=wf_setaper(&at_wf, nchan, diam, eps_inn, eps_out))<0) return err ;
	nw = wf_getnw(&at_wf) ;

	/* Compute the total number of weight functions that may be in use: */
	at_nweight[0] = nw ; /* Normal mode: one z-shift */
	
	at_nweight[1] = nw ; /* Generalized mode: two z-shifts */
	for (i=0; (i<(int)strlen(useshft)) && (i<nw); i++) 
		if (useshft[i]!='0') (at_nweight[1])++ ;

	/* Check that the total number of weights will be less than the number of
	altitude steps in orthogonal matrix: */
	if (at_nweight[1]>=nz) {
		at_done() ;
		return SC_ERZGR ;
	}

	/* Create the combined weight matrix for Cn2-restoration with 1km-cell grid.
	It works for both non-shifted mode (its first nw rows only) and for
	generalized mode (full matrix). */
	/* Number of z-steps in restoration grid: */
	nz = (int)(ceil((floor(zmax)-floor(z0))/AT_DZCN2+1)) ;
	if ((at_wcomb=dmatrix(0,at_nweight[1]-1,0,nz-1))==NULL) 
		return SC_ERMEM ;

	/* Creation of the matrices of decomposition coefficients */
	at_coef = (double***)malloc(AT_NMODE*sizeof(double**)) ;
	for (i=0; i<AT_NMODE; i++)
		if ((at_coef[i] = dmatrix(0,AT_NPOWER-1,0,at_nweight[i]-1))<0) 
			return SC_ERMEM ;	
		
	at_nchan = nchan ;

	return 0;
}

/*---------------------------- at_sedcrv ---------------------------------*/
/**
	@brief	Multiply the spectrum curve by the loaded response curve with 
		optionally shifted origins
	@param	lambda_sed	array of wavelengths of spectrum
	@param	sed  	array of spectrum
	@param	nsed	length of \em sed[] and \em lambda_sed[]
	@return	0 if Ok, <0 is errors (NULL pointers or if the spectrum wavelength 
	grid does not include the response wavelength grid.
	
	Function looks for the coincidence of some \em k-th element of \em 
	lambda_sed[] grid with the first element of the response curve wavelength
	grid \em lambda_crv. This \em k is a shift (or difference) of origins of
	SED and response curves (normally, the spectrum is wider). 

	Then, first elements of spectrum - \em sed[i], i=0..at_nresponse - are
	rewritten with the product of SED and response curves 
	\em	sed[i+k]*response[i] which takes into account the different origins of
	these functions. Last \em nsed-at_nresponse elements of \em sed[] are replaced with
	zeros. Thus, on exit only first \em at_nresponse elements of \em sed[] make
	sense.
	
	@see at_calcwf() at_checksed()
*/
int at_sedcrv(double * lambda_sed, double * sed, int nsed) {
	int k, j ;

	if ((lambda_sed==NULL) || (sed==NULL) || (nsed<=0)) return SC_ERNUL ;
	
	/* Find the starting point of the response curve at SED wavelength 
		grid: */
	k = -1 ;	/* Set "no match" */
	for (j=0; j<=nsed-at_nresponse; j++) 
		if (at_lambda[0]==lambda_sed[j]) { 
			k = j ; break; 
		}
	if (k<0) 
		return SC_ERCRV ;

	/* Multiply the arrays, given their different origin */
	for (j=0; j<at_nresponse; j++)
		/* Check the match of grids: */
		if (fabs(lambda_sed[j+k] - at_lambda[j])>WF_EPSLEFF) 
			return SC_ERCRV ;
		else 
			/* k is not negative: no data loss by overwrite below */
			sed[j] = sed[j+k] * at_response[j] ;
/*	for (j=0; j<at_nresponse; j++)
		printf("%5.3f %6.4f\n",at_lambda[j],sed[j]); */
	return 0 ;
}

/*---------------------------- at_loadsed --------------------------------*/

int at_loadsed(const char * sedfile) {
	double * lambda = NULL, * sed = NULL ;
	int err, nsed ;
	
	/* Check if the response curve is already loaded */
	if ((at_nresponse<=0) || (at_lambda==NULL))
		return SC_ERNUL ;
	
	/* Check that spectrum that we read from file is Ok: */
	if ((err=at_read2col(sedfile, &lambda, &sed, &nsed))!=0) return err ;

	/* Convert wavelengths into [mkm] */
	err=at_conwgrid(lambda, nsed, MKM_RANGE) ;
	
	/* Check that response curve wavelength grid complies the spectrum and
	multiply them: */
	if (!err)
		err = at_sedcrv(lambda, sed, nsed) ;

	/* We do not need SED's wavelength grid since it matches the response grid*/	
	free_dvector(lambda,0,nsed-1) ;
	if (err) {
		/* Release test-spectrum and lambda-grid: */
		free_dvector(sed,0,nsed-1) ;
	} else {
		/* Update the global storage replacing the current SED with a new one */
		if (at_sed)
			free_dvector(at_sed, 0, at_nsed) ;
		at_sed = sed ;
		at_nsed = nsed ;
	}
	return err ;
}

/*------------------------------ at_checkwf -------------------------------*/

int at_checkwf(const char * wfile) {
	int err ;
	if (wfile==NULL) return SC_ERNUL ;
	
	/* Compute and store the current spectrum characteristics in at_wf: */
	if ((err = wf_leff(&at_wf, at_lambda, at_sed, at_nresponse, 1))<0) 
		return err ;
	/* Check if the weight file wfile fits the current state of at_wf: */
	err = wf_checkfile(&at_wf, wfile) ;
	return err ;
}

/*------------------------------ at_calcwf --------------------------------*/

int at_calcwf(const char * wfile, void(* progress)(int)) {
	int err ;
	
	err = at_checkwf(wfile) ;
	if (!err) return 0 ;
	
	if (err!=SC_ERZGR) {
		/* Problem was not in the altitude grid: compute with
		default optimal grid : */
		wf_clear(&at_wf) ;
		err = wf_setzgrid(&at_wf, WF_Z0, WF_ZMAX, WF_DZ, WF_DZMIN) ;
		if (err>0) err = 0 ;	/* err was = nz */
	} /* else - Leave the user's grid intact */
	/* Since sed is shifted to the origin of response curve, Nsed =
			at_nresponse : */
	if (!err)
		err = wf_calc(&at_wf, at_lambda, at_sed, at_nresponse, progress) ;
	if (!err)
		/* Write the new weight file: */
		err = wf_write(&at_wf, wfile) ;

	return err ;
}

/*------------------------------ at_update --------------------------------*/

int at_update(
	const char * wfile,
	const char * checkfile,
	double zshift1,
	double zshift2) {
#define KM2M (1e3)
#define UPDRETURN(a) {free_dvector(lambda,0,nsed-1);free_dvector(sed,0,nsed-1);\
free(at_curwffile);at_curwffile=NULL;return (a);} 

	int i, j, k, nz, nbase, ibound ;
	int err = 0 ;
	int gen = 0 ;
	at_ipow ipow ;
	double d, power, curzshift ;
	/* SVD storages: */
	double ** wmatr = NULL, ** wcopy = NULL ;
	double * wvect = NULL ;
	double ** vmatr = NULL ;
	/* Altitude's powers vector */
	double * hvect = NULL ;
	const char useshft[] = AT_USESHFT ;
	wf_t at_wfs ;	/* Work weight structure */
	FILE * cf = NULL ;
	
	/* User must put the SED and weight file names to read. Although they may
	be the same as in previous run and cause no action at all */
	if (wfile==NULL) return SC_ERNUL ;
	
	/* Check the weight to coincide with the currently loaded by its file name */
	if ((at_curwffile==NULL) || strcmp(at_curwffile, wfile)) {
		/* weight is either not loaded or different */
		if (at_curwffile!=NULL) free(at_curwffile) ;
		at_curwffile = (char*)malloc(strlen(wfile)*sizeof(char)+1) ;
		strcpy(at_curwffile, wfile) ;
		/* Compute the response*SED in the SED's place: */

		/* Check and read (if ok) the file with weight: */
		err = at_checkwf(wfile) ;
		if (!err)
			err = wf_read(&at_wf, wfile) ;
		if (err<0) 
			return err ;
	} else {
		/* Weight is the same. Check for the shifts and exit if they are so: */
		if ((at_zshift[0]==zshift1) & (at_zshift[1]==zshift2))
			return 0 ;
	}
	
	/* Either weight functions or shifts have changed: redo decomposition */

	/* Cycle with 2 steps: 0-th for moments decomposition for the normal mode;
	   1-st for generalized (2 zshifts) mode */

	/* Check-file opeing: */
	if (checkfile!=NULL)
		if ((cf=fopen(checkfile,"wt"))==NULL) return SC_ERFIO ;

	wf_init(&at_wfs) ;
	for (gen=0; gen<AT_NMODE; gen++) {

		if (gen)
			curzshift = zshift2 ;
		else
			curzshift = zshift1 ;
		at_zshift[gen] = curzshift ;


		/* COMPUTE THE COMBINED WEIGHT MATRIX FOR Cn2 RESTORATION with dz=1km */

		/* (gen+1)-th part of this matrix is obtained now: */
		if ((err = wf_setzgrid(&at_wfs,floor(at_z0),floor(at_zmax),AT_DZCN2,0))<0) 
			return err;
		if ((err = wf_intowf(&at_wf, &at_wfs, curzshift, NULL, NULL, NULL))<0)
			return err;

		/* Fill in the respective part of combined weight matrix */
		if (cf!=NULL) fprintf(cf,"# Use weights:") ;
		if (gen) fprintf(cf," non-shifted +");
		for (j=0, k=gen*at_nweight[0]; k<at_nweight[gen]; j++) {
			if ((!gen) || (useshft[j]!='0')) {
				wf_copy(&at_wfs,0, wf_getnz(&at_wfs)-1, j, at_wcomb[k]) ;
				if (cf!=NULL) {
					if (!gen)
						fprintf(cf," %s",wf_wname(&at_wfs,j)) ;
					else
						fprintf(cf," %ss",wf_wname(&at_wfs,j)) ;
				}
				k++ ;
			}
		}
		
		if (cf!=NULL) fprintf(cf, "\n") ;
		/* Clear shifted weight back: */
		wf_clear(&at_wfs) ;
		wf_setaper(&at_wfs,0,0,NULL,NULL) ;

		/* COMPUTE THE TRANSPOSED WEIGHT MATRIX FOR ALTITUDE POWERS
		DECOMPOSITION */
		if ((err = wf_setzgrid(&at_wfs,at_z0,at_zmax,at_dz,at_dzmin))<0) 
			return err;
		if ((err = wf_intowf(&at_wf, &at_wfs, curzshift, NULL, NULL, NULL))<0)
			return err;
		
		if (!gen) {
			/* First loop: allocate memory */
			nz = wf_getnz(&at_wfs) ;
			if ((wmatr=dmatrix(1,nz,1,at_nweight[1]))==NULL) return SC_ERMEM;
			if ((wcopy=dmatrix(1,nz,1,at_nweight[1]))==NULL) return SC_ERMEM;
			if ((wvect=dvector(1,at_nweight[1]))==NULL) return SC_ERMEM ;
			if ((vmatr=dmatrix(1,at_nweight[1],1,at_nweight[1]))==NULL) 
				return SC_ERMEM ;
			if ((hvect=dvector(0,nz-1))==NULL) return SC_ERMEM ;

			/* Find the zgrid-index of the boundary layer end: */
			ibound = 0 ;
			for (j=0; j<nz; j++) {
				if (wf_getalt(&at_wfs,j)>=AT_HBOUND) {
					ibound = j;
					break ;
				}
			}
			if (ibound==0) return SC_ERZGR ;
		} else {
			/* Restore the wmatr from the copy to update its right-side col-s */
			for (j=0; j<at_nweight[0]; j++) 
				for (i=0; i<nz; i++) 
					wmatr[i+1][j+1] = wcopy[i+1][j+1] ;
		}
	
		/* Combined transposed matrix of weights for SVD */
		for (j=0,k=gen*at_nweight[0]; k<at_nweight[gen]; j++) {
			if ((!gen) || (useshft[j]!='0')) {
				for (i=0; i<nz; i++) 
					wmatr[i+1][k+1] = wcopy[i+1][k+1] = wf_getval(&at_wfs,i,j) ;
				k++ ;
			}
		}

		/* Checking for permutation invariancy: */
/*		for (j=0;j<nz;j++) { d = wmatr[j+1][5];	wmatr[j+1][5] = wmatr[j+1][16] ;
			wmatr[j+1][16] = d;	}*/
	
		/* Run SVD of weight matrix: */
		svdcmp(wmatr,nz,at_nweight[gen],wvect,vmatr) ;

		/* scalar product check: */
/*		for (i=0;i<at_nweight[gen]-1;i++) {
			d = 0; for (j=0;j<nz;j++) d+= wmatr[j+1][i+1]*wmatr[j+1][i+2] ;
			printf("%d-%d %10.4e\n",i,i+1,d); }
*/
		/* Maximal singular value: */
		for (d=0,i=1;i<=at_nweight[gen];i++) 
			if (wvect[i]>d) d=wvect[i] ; /* maximal w */
		/* Minimal acceptable singular value: */
		d *= AT_MINRELW ;	

		nbase = at_nweight[gen] ;
		/* Rejection of small singulars: */
		if (cf!=NULL) fprintf(cf, "# N(Base functions): Init.=%d, ",nbase) ;
		for (i=1;i<=at_nweight[gen];i++) 
			if (wvect[i]<d) { 
				wvect[i] = 0 ; 
				nbase--; 
			}
		if (cf!=NULL) fprintf(cf, " Smalls reject=> %d, ",nbase);
		/* Rejection of least singulars: provided they are at the end of 
			wvect[] */
		for (i=at_nweight[gen];(i>0) && (nbase>AT_MAXNBASE);i--) 
			if (wvect[i]!=0) {
				wvect[i] = 0 ; 
				nbase--; 
			}
		if (cf!=NULL) 
			fprintf(cf, " Restricted to => %d\n# Approximated powers: ",nbase);
	
		for (ipow=AT_IFSE; ipow<=AT_IISK; ipow=(at_ipow)((int)ipow+1)) {
			/* Select the power: */
			switch (ipow) {
				case AT_IFSE:
				case AT_IWSE: power = AT_POWSEE; break ;
				case AT_IISP: power = AT_POWISP; break ;
				case AT_IEFF: power = AT_POWEFF; break ;
				case AT_IISK: power = AT_POWISK; break ;
				default: return SC_ERPAR ;				
			}
			if (cf) fprintf(cf, "\n# Ap%d: %6.3f",(int)ipow,power) ;
			/* Generate the altitude powers: */
			for (j=0; j<nz; j++) 
				hvect[j] = pow(wf_getalt(&at_wfs,j)*KM2M, power) ;
			if (ipow==AT_IFSE) {
				/* free atmosphere: force the hvect to resemble the smallest
				aperture pair weight behavior: */
				wf_copy(&at_wfs,0,ibound-1,at_nchan,hvect) ;
				d = wf_getval(&at_wfs,ibound,at_nchan) ;
				for (j=0; j<ibound; j++) hvect[j] /= d ;
				if (cf) fprintf(cf," (W(AB)-dumped)");
			}
			/* Check the powers: */
/*if (ipow==AT_IFSE) for (j=0;j<nz;j++) printf("%f %f\n",wf_getalt(&at_wfs,j),hvect[j]);*/
			/* Back-substitute to find the coefficients: */
			svbksb(wmatr,wvect,vmatr,nz,at_nweight[gen], hvect-1, 
				at_coef[gen][(int)ipow]-1);
		
		}
		if (cf!=NULL) {
			/* Check the coefficients: */
			fprintf(cf,"\n# Coefficients:\n");
			for (ipow=AT_IFSE; ipow<=AT_IISK; ipow=(at_ipow)((int)ipow+1)) {
				/* Maximal one: */
				for (d=0,j=0; j<at_nweight[gen]; j++) 
					if (d<at_coef[gen][(int)ipow][j]) 
						d=at_coef[gen][(int)ipow][j];
				fprintf(cf,"# Ap%d: Max=%9.3e",(int)ipow,d);
				for (j=0; j<at_nweight[gen]; j++) 
					fprintf(cf," %7.4f",at_coef[gen][(int)ipow][j]/d); 
				fprintf(cf,"\n");
			}
			/* Check the approximated relations: */	
			fprintf(cf,"# h[km]  ") ;
			for (i=0; i<=(int)AT_IISK ; i++) 
				fprintf(cf,"#Ap%d/h^ ",i) ;
			fprintf(cf,"\n") ;
			for (i=1;i<nz;i++) {
				fprintf(cf,"%6.3f",wf_getalt(&at_wfs,i)) ;
				for (ipow=AT_IFSE; ipow<=AT_IISK; ipow=(at_ipow)((int)ipow+1)) {
					switch (ipow) {
						case AT_IFSE:
						case AT_IWSE: power = AT_POWSEE; break ;
						case AT_IISP: power = AT_POWISP; break ;
						case AT_IEFF: power = AT_POWEFF; break ;
						case AT_IISK: power = AT_POWISK; break ;
						default: return SC_ERPAR ;				
					}
					/* power of altitude: */
					curzshift = pow(wf_getalt(&at_wfs,i)*KM2M,power) ;
					for (d=0,j=0; j<at_nweight[gen]; j++) 
						d+=wcopy[i+1][j+1]*at_coef[gen][ipow][j];
					fprintf(cf," %6.3f",d/curzshift); 
				}
				fprintf(cf,"\n") ;
			}
		}
		/* Clear the shifted weight */
		wf_clear(&at_wfs) ;
		wf_setaper(&at_wfs,0,0,NULL,NULL) ;
	}

	/* Close check-file */
	if (checkfile!=NULL)
		fclose(cf) ;

	/* Free the orthogonalized matrix, singular value vector and square 
	V-matrix: */
	free_dmatrix(wmatr,1,nz,1,at_nweight[1]) ;
	free_dmatrix(wcopy,1,nz,1,at_nweight[1]) ;
	free_dvector(wvect,1,at_nweight[1]) ;
	free_dmatrix(vmatr,1,at_nweight[1],1,at_nweight[1]) ;
	free_dvector(hvect,0,nz-1) ;

	return 0;
#undef KM2M
}

/*------------------------------ at_alloc --------------------------------*/

int at_alloc(int nmeas) {
	/* Allocate the storages for integrals: */
	if (nmeas>at_maxnmeas) {
		if (at_maxnmeas>0) {
			free_dmatrix(at_int,0,at_maxnmeas-1,0,AT_LAST-1) ;
			free_dvector(at_aint,0,AT_LAST-1) ;
			free_dvector(at_e2int,0,AT_LAST-1) ;
		}
		if ((at_int=dmatrix(0,nmeas-1,0,AT_LAST-1))==NULL) return SC_ERMEM ;
		if ((at_aint=dvector(0,AT_LAST-1))==NULL) return SC_ERMEM ;
		if ((at_e2int=dvector(0,AT_LAST-1))==NULL) return SC_ERMEM ;
		at_maxnmeas = nmeas ;
	}
	return 0 ;
}

/*------------------------------ at_done --------------------------------*/

void at_done(void) {
	if (at_curwffile!=NULL)
		free(at_curwffile) ;
		
	if (at_sed!=NULL) {
		free_dvector(at_sed,0,at_nsed) ;
		at_sed = NULL ;
		at_nsed = 0 ;
	}
	
	if (at_nresponse>0) {
		free_dvector(at_response,0,at_nresponse-1);
		free_dvector(at_lambda,0,at_nresponse-1);
		at_nresponse = 0 ;
		at_response = at_lambda = NULL ;
	}
	if ((at_nweight[1]>0) && (at_wcomb!=NULL)) {
		free_dmatrix(at_wcomb,0,at_nweight[1]-1,0,0) ;
		at_wcomb = NULL ;
	}
	if (at_ncn2>0) {
		free_dvector(at_cn2,0,at_ncn2);
		free_dvector(at_zcn2,0,at_ncn2);
		at_ncn2 = 0 ;
		at_cn2 = at_zcn2 = NULL ;
	}
	wf_clear(&at_wf) ;
	wf_setaper(&at_wf,0,0,NULL,NULL) ;
	at_nchan = 0 ;
	
	if (at_coef!=NULL){ 
		free_dmatrix(at_coef[0],0,AT_NPOWER-1,0,AT_NPOWER); /* last par. is foo*/
		free_dmatrix(at_coef[1],0,AT_NPOWER-1,0,AT_NPOWER) ;
		free(at_coef) ;
		at_coef = NULL ;
		at_nweight[0] = at_nweight[1] = 0 ;
	}

	/* Deallocate the storages for integrals: */
	if (at_int!=NULL) {
		free_dmatrix(at_int,0,at_maxnmeas-1,0,AT_LAST-1) ;
		free_dvector(at_aint,0,AT_LAST-1) ;
		free_dvector(at_e2int,0,AT_LAST-1) ;
		at_int = NULL ; at_aint = at_e2int = NULL ;
		at_maxnmeas = -1 ;
	}

	return ;
}

/*------------------------------ at_calcint --------------------------------*/

int at_calcint(double * scind, double * scinds, int meas) {

	int i,j,k,isgen ;
	double cn2mom ;
	const char useshft[] = AT_USESHFT ;

	if ((at_coef==NULL) || (scind==NULL)) return SC_ERNUL ;

	/* We detect the generalized mode by the appearance of non-NULL scinds */
	isgen = (scinds!=NULL) ;

	/* Compute the Cn2-integrals from indices: */
	for (i=0; i<AT_NPOWER; i++) {
		for (j=0,cn2mom=0; j<at_nweight[0]; j++) 
		 	cn2mom += scind[j] * at_coef[isgen][i][j] ;
		if (isgen) {
			for (j=0,k=at_nweight[0]; k<at_nweight[1]; j++) 
				if (useshft[j]!='0') {
				 	cn2mom += scinds[j] * at_coef[isgen][i][k] ;
					k++ ;
				}
		}
		at_int[meas][i] = cn2mom ;
	}
	return 0 ;
}


/*---------------------------- at_meritfix --------------------------------*/
/**
	@brief	Merit function for the profile-restoration at fixed altitudes
	@param	y	vector of sqrt(Cn2) of layers
	@return	merit Chi-square
	
*/
static int at_curnweight = 0 ;

static double at_meritfix(double * y) {
	register int i, j ;
	double d, ch2, s2 ;

	for (ch2=0,j=0; j<at_curnweight; j++) {
		for (s2=0,i=0; i<at_ncn2; i++) 
			s2 += y[i+1] * y[i+1] * at_wcomb[j][(int)(at_zcn2[i]+0.5)] ;
		d = s2 - at_sicomb[j] ;
		ch2 += d*d / at_e2sicomb[j] ;
	}
	return ch2 ;
}

/*---------------------------- at_calcn2 ----------------------------------*/

double at_calcn2(double * scind, double * scinds,
		double * e2scind, double * e2scinds, int isfixed) {
#define NFIXED 6	/* Number of fitted layers with fixed altitude */
#define NFLOAT 4	/* ditto with adjusted altitudes */
#define ZFIXFACT 2.0	/* Factor of altitude modification in fixed-altitude method,
should be about 1/(alt.resolution) */
#define ZGND 0
#define ZBOUND 1
#define AVGCN2	(6.95e-13)	/* Cn2-integral giving 1" seeing */
#define FTOL	(1e-3)		/* Tolerance in powell() */

#define MAXH1 5 	/* Maximal altitude of the second floating layer */
#define MAXH2 12	/* Maximal altitude of the third floating layer */
#define MINDH 0		/* Minimal alitude difference between layers */
	int i, j, k, iter=0 ;
	const char useshft[] = AT_USESHFT ;
	double ** xi = NULL ;
	double chi2 = -1, minchi2, d, s2 ;
	int ih[NFLOAT] ;	/* Trial altitudes of layers */
	double cn2[NFLOAT] ;	/* ... and their strengths */
	double ** w3, ** w3inv ;	/* small weight matrix and its inversion */
	double * wvect, ** vmatr ;

	/* Combine the indices: */
	if (scinds==NULL) 
		at_curnweight = at_nweight[0] ;
	else
		at_curnweight = at_nweight[1] ;

	if ((at_sicomb=dvector(0,at_curnweight-1))==NULL) return SC_ERMEM ;
	if ((at_e2sicomb=dvector(0,at_curnweight-1))==NULL) return SC_ERMEM ;
	for (i=0; i<at_nweight[0]; i++) {
		at_sicomb[i] = scind[i] ;
		at_e2sicomb[i] = e2scind[i] ;
	}
	for (i=0,k=at_nweight[0]; k<at_curnweight; i++) {
		if (useshft[i]!='0') {
			at_sicomb[k] = scinds[i] ;
			at_e2sicomb[k] = e2scinds[i] ;
			k++ ;
		}
	}

	/* Allocate maximal of possibly needed memory if not allocated */
	i = (NFIXED>NFLOAT ? NFIXED : NFLOAT) ;
	if (at_cn2==NULL) {
		/* Allocate maximum needed vector +1 element for Chi-2 in at_cn2 */
		if ((at_zcn2=dvector(0,i-1))==NULL) return SC_ERMEM ;
		if ((at_cn2=dvector(0,i))==NULL) return SC_ERMEM ;
	}

	if (isfixed) {
		/* Fixed layers method */
		/* Generate the grids: altitudes and 1st approximation of the profile */
		at_ncn2 = NFIXED ;
		at_zcn2[0] = ZGND + (double)(scinds==NULL);	/* ground layer */
		at_zcn2[1] = ZBOUND ;
		for (i=2; i<at_ncn2; i++) {
			at_zcn2[i] = at_zcn2[i-1]*ZFIXFACT ;
			at_cn2[i] = sqrt(AVGCN2 / at_ncn2) ;
		}
		/* Diagonal "identity" matrix: */
		if ((xi=dmatrix(1,at_ncn2,1,at_ncn2))==NULL) return SC_ERMEM ;
		for (i=1; i<=at_ncn2; i++) 
			for (j=1; j<=at_ncn2; j++) xi[i][j] = (double)(i==j) ;
			
		/* Solve the task: find the minimum chi-square */
		powell(at_cn2-1,xi,at_ncn2,FTOL,&iter,&minchi2,at_meritfix);
		/* Back-square the profile: */
		for (i=0; i<at_ncn2; i++) at_cn2[i] *= at_cn2[i] ;
		at_cn2[at_ncn2] = minchi2 ;
		free_dmatrix(xi,1,at_ncn2,1,at_ncn2) ;
	} else {
		/* "Floating" layers method */
		at_ncn2 = NFLOAT ;
		if ((w3=dmatrix(1,at_curnweight,1,at_ncn2))==NULL ) return SC_ERMEM ;
		if ((w3inv=dmatrix(1,at_ncn2,1,at_curnweight))==NULL ) return SC_ERMEM ;
		if ((wvect=dvector(1,at_ncn2))==NULL) return SC_ERMEM ;
		if ((vmatr=dmatrix(1,at_ncn2,1,at_ncn2))==NULL) return SC_ERMEM ;
		ih[0] = (scinds==NULL) ; /* Fixed ground layer */
		minchi2 = 1e15 ;
		for (ih[1]=ih[0]+MINDH; ih[1]<=MAXH1; ih[1]++)
			for (ih[2]=ih[1]+MINDH; ih[2]<=MAXH2; ih[2]++)
				for (ih[3]=ih[2]+MINDH; ih[3]<=(int)at_wf.zmax; ih[3]++) {
					/* Compose the selected weights matrix */
					for (j=0;j<at_curnweight;j++)
						for (i=0;i<at_ncn2; i++)
							w3[j+1][i+1] = at_wcomb[j][ih[i]] ;

					/* Invert w3-matrix */
					svdcmp(w3,at_curnweight,at_ncn2,wvect,vmatr) ;
					/* Cancel small singulars */
					for (i=1,d=0; i<=at_ncn2; i++)
						if (d<wvect[i]) d=wvect[i];
					d *= AT_MINRELW ;	/* Minimal acceptable singular value */
					for (i=1; i<=at_ncn2; i++)
						if (wvect[i]>d) 
							wvect[i] = 1. / wvect[i] ;
						else 
							wvect[i] = 0 ;
					/* multiply three matrices: A^{-1}=(V*diag(1/w))*U^T */
					for (i=1; i<=at_ncn2; i++) 
						for (j=1; j<=at_curnweight; j++)
							for (k=1,w3inv[i][j]=0; k<=at_ncn2; k++)
								w3inv[i][j] += vmatr[i][k]*wvect[k]*w3[j][k] ;

/* Check the inversion: reproduce and display w3, display w3inv and their
product: */
/*
				if (ih[1]==5 && ih[2]==12 && ih[3]==20) {
					for (j=0;j<at_curnweight;j++) {
						for (i=0;i<at_ncn2; i++) {
							w3[j+1][i+1] = at_wcomb[j][ih[i]] ;
							printf("%6.3f ",1e-11*w3[j+1][i+1]);
						}
						printf("\n");
					}
					printf("\n");
					for (i=0;i<at_ncn2; i++) {
						for (j=0;j<at_curnweight;j++) {
							printf("%6.3f ",1e+11*w3inv[i+1][j+1]);
						}
						printf("\n");
					}
					printf("Ok\n");
					for (i=1;i<=at_ncn2; i++) {
						for (j=1; j<=at_ncn2; j++) {
							for (k=1,d=0; k<=at_curnweight; k++)
								d+=w3inv[i][k] * w3[k][j];
							printf(" %5.2f",d);
						}
						printf("\n");
					}
				}
*/
					/* Compute the strengths: */
					for (i=0; i<at_ncn2; i++) {
						for (cn2[i]=0,j=0; j<at_curnweight; j++)
							cn2[i] += w3inv[i+1][j+1] * at_sicomb[j] ;
						if (cn2[i]<0) cn2[i] = 0 ;	  /* >=0 constraint */
					}

					/* Compute indices and chi-square: */
					for (chi2=0,j=0; j<at_curnweight; j++) {
						for (s2=0,i=0; i<at_ncn2; i++)
							s2 += cn2[i] * at_wcomb[j][ih[i]] ;
						d = s2 - at_sicomb[j];
						chi2 += d*d / at_e2sicomb[j] ;
					}
					/* Save the altitudes and strengths if this is the best
						case up to now */
					if (chi2<minchi2) {
						minchi2 = chi2 ;
						for (i=0; i<at_ncn2; i++) {
							at_zcn2[i] = (double)(ih[i]) ;
							at_cn2[i] = cn2[i] ;
						}
					}

				}
		free_dmatrix(w3,1,at_curnweight,1,at_ncn2) ;
		free_dmatrix(w3inv,1,at_ncn2,1,at_curnweight) ;
		free_dvector(wvect,1,at_ncn2) ;
		free_dmatrix(vmatr,1,at_ncn2,1,at_ncn2) ;
		at_cn2[at_ncn2] = minchi2 ;
	}
	free_dvector(at_sicomb,0,at_curnweight-1) ;
	free_dvector(at_e2sicomb,0,at_curnweight-1) ;
	return minchi2 ;

#undef NFIXED
#undef NFLOAT
#undef ZFIXFACT
#undef ZGND
#undef ZBOUND
#undef AVGCN2
#undef FTOL
#undef MAXH1
#undef MAXH2
}

/*------------------------------ at_avgints ----------------------------------*/

/** Compute the formal relative error of the value of the power function of the
value \em mean given its squared error \em e2 */
static double at_relerr(double mean, double e2, double power) {
	if ((e2<=0) || (mean<=0)) 
		return 0 ;
	else
		return fabs(power * sqrt(e2) / mean) ;
}

int at_avgint(int nmeas, double desi, double e2desi, double zdist) {
#define JPOW_SEE	(3./5.)	/* Power of Cn-integral for seeing */
#define LPOW_SEE	(-1./5.)	/* Power of lambda for seeing */
#define ZPOW_SEE	(-1.)	/* secans of zenith distance power for seeing */
#define JPOW_ISP	(-3./5.)	/* Power of Cn-5/3th moment for isoplan. angle*/
#define LPOW_ISP	(6./5.)	/* Power of lambda for isoplan. angle */
#define ZPOW_ISP	(8./5.)	/* secans of zenith distance power for isop.angle */
#define SPOW_TC  	(-3./5.) /* Power of DESI for atm. time constant */
	int err ;
	double d, power, zfact ;

	if (nmeas<1) 
		return SC_ERNOD ;
	if (nmeas>at_maxnmeas)
		return SC_EROFL ;

	/* average integrals: */
	err = sc_avgmatrix(at_int,AT_LAST,nmeas,NULL,0,at_aint,at_e2int,SC_MAXLAG) ;
	if (err<0) return err ;

	/* Convert integrals into atmospheric parameters */
	
	/* secans of the zenith distance [deg] */
	zfact = 1./cos(zdist*M_PI/180.) ;

	/* Seeing : free and whole atmosphere */	
	at_fseeing = AT_KS * pow(at_aint[AT_IFSE],JPOW_SEE) * 
		pow(at_wf.leff,LPOW_SEE) * pow(zfact,ZPOW_SEE) ;
	at_efseeing = at_fseeing * 
		at_relerr(at_aint[AT_IFSE],at_e2int[AT_IFSE],JPOW_SEE);

	at_seeing = AT_KS * pow(at_aint[AT_IWSE],JPOW_SEE) * 
		pow(at_wf.leff,LPOW_SEE) * pow(zfact,ZPOW_SEE) ;
	at_eseeing = at_seeing * 
		at_relerr(at_aint[AT_IWSE],at_e2int[AT_IWSE],JPOW_SEE);
	
	/* Isoplanatic angle: */
	at_isoplan = AT_KP * pow(at_aint[AT_IISP],JPOW_ISP) * pow(zfact,ZPOW_ISP)
		* pow(at_wf.leff, LPOW_ISP) ;
	at_eisoplan = at_isoplan *
		at_relerr(at_aint[AT_IISP],at_e2int[AT_IISP],JPOW_ISP); 

	/* Effective altitude of turbulence in [meters]: (free atm.) */
	power = 1.0/(AT_POWISP-AT_POWEFF) ;
	at_fheff = pow(at_aint[AT_IISP]/at_aint[AT_IEFF],power) * 
				pow(zfact,(AT_POWISP-AT_POWEFF));
	d = at_relerr(at_aint[AT_IISP],at_e2int[AT_IISP],power) ;
	at_efheff = d*d;
	d = at_relerr(at_aint[AT_IEFF],at_e2int[AT_IEFF],power) ;
	at_efheff += d*d;
	at_efheff *= at_fheff ;
	
	/* Whole atm. effective altitude: */
	at_heff = at_aint[AT_IEFF]/at_aint[AT_IWSE] * zfact ;
	d = at_relerr(at_aint[AT_IEFF],at_e2int[AT_IEFF],AT_POWEFF) ;
	at_eheff = d*d;
	d = at_relerr(at_aint[AT_IWSE],at_e2int[AT_IWSE],AT_POWEFF) ;
	at_eheff += d*d;
	at_eheff *= at_heff ;

	at_timeconst = AT_KT * pow(desi,SPOW_TC) ;
	if (e2desi>0) d = sqrt(e2desi) ; else d = 0 ;
	at_etimeconst = at_timeconst * at_relerr(desi,e2desi,SPOW_TC) ;
	
	return 0 ;
#undef LENFMT
#undef JPOW_SEE
#undef LPOW_SEE
#undef ZPOW_SEE
#undef JPOW_ISP
#undef LPOW_ISP
#undef ZPOW_ISP
#undef SPOW_TC 
}

/*----------------------------- at_writeint -------------------------------*/

int at_writeint(FILE * f, const char * stamp) {
#define LENFMT 20	
	int i ;
	double d, e ;
	const char fmt[][LENFMT] = {AT_INTFMT} ;

	if (f==NULL) return SC_ERPAR ;
	
	/* write the stamp: */
	if (stamp!=NULL) {
		if (!fprintf(f, "%s",stamp)) return SC_ERFIO ;
	}
	
	/* write the average integrals and their errors: */
	for (i=0; i<AT_LAST; i++) {
		switch (i) {
			case AT_FSEE :
				d = at_fseeing; e = at_efseeing ; break ;
			case AT_WSEE :
				d = at_seeing; e = at_eseeing ; break ;
			case AT_ISP :
				d = at_isoplan; e = at_eisoplan ; break ;
			case AT_M2 :
				d = at_aint[AT_IISK]; e = at_e2int[AT_IISK]; 
				if (e>0) e = sqrt(e) ; else e = 0 ; 
					break ;
			case AT_FHEFF :
				d = at_fheff; e = at_efheff ; break ;
			case AT_HEFF :
				d = at_heff; e = at_eheff ; break ;
			case AT_TC :
				d = at_timeconst; e = at_etimeconst ; break ;
			default:
				return SC_ERPAR ;
		}
		fprintf(f,fmt[i],AT_INTDLM,d) ;
		fprintf(f,fmt[i],AT_INTDLM,e) ;
	}
	if (fprintf(f,"\n")==0) return SC_ERFIO ;
	return 0 ;
}

/*----------------------------- at_writecn2 -------------------------------*/

int at_writecn2(FILE * f, const char * stamp) {
#define LENFMT 20	
	int i ;

	if (f==NULL) return SC_ERPAR ;
	
	/* write the stamp: */
	if (stamp!=NULL) {
		if (!fprintf(f, "%s",stamp)) return SC_ERFIO ;
	}
	
	/* save the dimention, altitudes and strengths of the restored profile: */
	fprintf(f,"%s%d%s%6.2f",AT_INTDLM,at_getncn2(),AT_INTDLM,
		at_getcn2()[at_getncn2()]) ;
	for (i=0; i<at_getncn2(); i++)
		fprintf(f,AT_ALTFMT,AT_INTDLM,at_getzcn2()[i]) ;
	for (i=0; i<at_getncn2(); i++)
		fprintf(f,AT_CN2FMT,AT_INTDLM,at_getcn2()[i]) ;
	if (fprintf(f,"\n")==0) return SC_ERFIO ;
	return 0 ;
}


