/****************************************************************************/
/*                 MASS project: TURBINA module file                        */
/*                            scind.c                                       */
/*                    Description is given in scind.h                       */
/* Editor tabular size: 4                                                   */
/****************************************************************************/

/* Insert the definition SCINDTEST in compilation for the module self-tests:
gcc -c scind.c .... -DSCINDTEST=1 */
//#define SCINDTEST

#include <math.h>
#include "nr.h"
#include "nrutil.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "scind.h"

#ifdef SCINDTEST
/*---------------------------- TEST PART ---------------------------------*/
#define NTEST 1
/* Uncomment the definition SPEEDTEST to omit data re-simulation and printouts
to obtain the clean performance estimate: */
//#define SPEEDTEST

/*----- Comes normally from configuration file: cf_... vars and defs --------*/

/* Apertures: */
/* number of apertures in segmentor : */
#define CF_NCHAN  4
/* Physical size of segmentator annular apertures [cm] : */
double cf_aper_inn[] = { 0.000, 0.102, 0.188, 0.354};
double cf_aper_out[] = { 0.100, 0.185, 0.350, 0.650};
/* Magnification of the system: */
double cf_magnif = 20.0 ;

/* Exposure section: */
double cf_microexp = 0.001 ;	/* < microexposure: timespan per count */
#define CF_BASENUMBER 1000 		/* number of microexposures per series */
#define CF_NMEAS 100			/* Number of base times in Accum. Time */

/* Detectors section: */
const double cf_deadtime[] = { 30E-9, 30E-9, 30E-9, 30E-9 };
//const double cf_deadtime[] = { 30E-12, 30E-12, 30E-12, 30E-12 };
/* < given for exposure time EXACTLY 1 sec !!! Here - about 30 ns */
//const double cf_nonpois[] = {  0.984,  0.993,  1.001,  0.989 };    
const double cf_nonpois[] = {  1.0,  1.0,  1.0,  1.0 };    
/* background counts measured in channels per second */
//const double cf_bkgr[] = {30., 32., 40., 70.}; 
const double cf_bkgr[] = {0., 0., 0., 0.};

#define CF_IDXFILE "index.dat"
#define CF_MOMFILE "rawdat.dat"
#define CF_AVGIDXFILE "aindex.dat"

/*------------- Other supplementary info: ------------------------*/
double sys_qe = 0.05 ;	/* System quantum efficiency */
double flux0 = 1E6 ; 	/* Photons per second per cm^2 from the mag=0 A0-star*/
#define PI_NUMB (3.1415926535897932) 


int main( int ac, char **av ) {

/* Receiving buffers for channels counts: */
count_t channelA[CF_BASENUMBER] ;
count_t channelB[CF_BASENUMBER] ;
count_t channelC[CF_BASENUMBER] ;
count_t channelD[CF_BASENUMBER] ;
count_t * channel[] = { (count_t*)(&channelA), (count_t*)(&channelB), 
						(count_t*)(&channelC), (count_t*)(&channelD) } ;

double area, flux, magnitude=0, modul=0, mean, factor, minf, maxf, utc = 0.0 ;

FILE * fdat = NULL ;	/* External counts file to reduce */
FILE * fidx = NULL ;	/* Index writing file */
FILE * fmom = NULL ;	/* Moments writing file */
FILE * favg = NULL ;	/* Average indices writing file */
#define DATALINE	(100)
char 	buf[DATALINE] ;	/* char. buffer to use */
char * s ; 	/* running pointer to this buffer */
#define STAMPLEN (1+1+1+1+8+1)	/* "# M HH.HHHHH", M=Mode symb; HH.HHHHH = UTC
hour */ 
char stamp[STAMPLEN] ;

int i, j, idum2=(-13),err, test, meas, stattest;
double stind[CF_NCHAN] ;	/* Theoretical indices from statistical test */
long idum=(-13) ;
int ncorr[CF_NCHAN] = {0,0,0,0} ;

if (ac<2) {
	printf(" MASS project: SCIND module test\n");
	printf(" Use: %s <star magnitude> <modulation factor [0,1[> [STATTEST]\n",av[0]);
	printf(" or: \n");
	printf("       %s <datafile:  %d channel columns with Nx%d counts each>\n",
		av[0],CF_NCHAN, CF_BASENUMBER);
	return -1 ;
}

if ((fdat=fopen(av[1],"rt"))==NULL) {
	magnitude = atof(av[1]) ;
	if (ac>2) modul = atof(av[2]) ;
	else modul = 0 ;
	if ((magnitude<-2) || (fabs(modul)>1)) {
		printf(" Error: magnitude is >=-2; modul<=1.0\n");
		return -1 ;
	}
	stattest = (ac>3) ;
} else {
	/* File's open: read counts */
	printf("File %s reading mode\n",av[1]);
	stattest = 0 ;
}

if (stattest)
	printf("Statistical test of MASS activated\n");

for (test=0; test<NTEST; test++) {
/* Cycle of all the system work scenario repetitions : */

	printf("Initialize the module.\n") ;
	err = sc_init(CF_BASENUMBER, CF_NMEAS, CF_NCHAN); 
	if (err) { fprintf(stderr,sc_error(err)); return err;}
	
	printf("Initialize Channels.\n") ;
	for (i=0; i<CF_NCHAN; i++) {
		sc_chan_init(i, channel[i], 
			cf_nonpois[i], 
			cf_bkgr[i] * cf_microexp,
			cf_deadtime[i] / cf_microexp ) ; 
//		printf("chan[%d].data => %x\n",i,(int)channel[i]);
	}

	printf("Opening the index output files %s and %s\n",
		CF_IDXFILE, CF_AVGIDXFILE);
	if ((fidx=fopen(CF_IDXFILE,"at"))==NULL) {
		fprintf(stderr,"File %s open error\n",CF_IDXFILE);
		return SC_ERFIO ;
	} else {
		/* write header: */
		if ((err=sc_writeidx(fidx,SC_HEAD,"# Mode UTC"))<0) {
			fprintf(stderr,"File %s header writing error\n",CF_IDXFILE);
			return err ;
		}
	}
	if ((favg=fopen(CF_AVGIDXFILE,"at"))==NULL) {
		fprintf(stderr,"File %s open error\n",CF_IDXFILE);
		return SC_ERFIO ;
	}	

	printf("Opening the moments output file %s\n",CF_MOMFILE);
	if ((fmom=fopen(CF_MOMFILE,"at"))==NULL) {
		fprintf(stderr,"File %s open error\n",CF_MOMFILE);
		return SC_ERFIO ;
	} else {
		/* write header: */
		if ((err=sc_writemom(fmom,SC_HEAD,"# Mode UTC"))<0) {
			fprintf(stderr,"File %s header writing error\n",CF_MOMFILE);
			return err ;
		}
	}

	for (meas=0; meas<CF_NMEAS; meas++) {
		if (fdat==NULL) {
#ifdef SPEEDTEST
			if (meas==0) 
#endif			
			{
			printf("\nUT=%7.4fh: Simulate data in channels:\n", utc);
			for (i=0; i<CF_NCHAN; i++) {
				/* Uncomment following idum-initialization to make the
				scintillations themselves similar in all channels but leaving photon
				statistics independent (use of ran1() and ran2() ! ): */
//				idum = -13;
				maxf = cf_magnif*cf_aper_out[i]/2 ; 
				minf = cf_magnif*cf_aper_inn[i]/2 ;
				area = (maxf*maxf - minf*minf) * PI_NUMB ;
				flux = flux0 * area * pow(10.0,-0.4*magnitude) * sys_qe ;
				printf ("%s: %5.2f cm2, flux=%10.1f ph/s =>", sc_apername(CF_NCHAN,i),
						area,flux) ;
				mean = maxf = 0 ; minf = 1e9 ;
				for (j=0; j<CF_BASENUMBER; j++) {
					if (stattest) {
						if (!(j%4-1))
							factor = 1.0 + modul ;
						else {
							if (!(j%4-3))
								factor = 1.0 - modul ;
							else
								factor = 1.0 ;
						}
					} else {
						factor = 1.0 + modul*(2.0*ran2(&idum) - 1.0);
					}
					if (factor<minf) minf=factor;
					else if (factor>maxf) maxf=factor;

					channel[i][j] =	(count_t)(0.5+
						poidev((flux*factor+cf_bkgr[i])*cf_microexp, &idum2)) ;
					mean += (double)channel[i][j] ;
//					printf("%d ",channel[i][j]);
//					channel[i][j] = channel[0][j];
				}
				mean /= CF_BASENUMBER ;
				printf(" Mean count: %8.1f (modul:%4.1f%%-%5.1f%%)\n",mean,
				minf*100, maxf*100);
			}
			}
		} else {
			printf("Read data series from %s:\n",av[1]);
			for (j=0; j<CF_BASENUMBER; j++) {
				if (fgets(buf,DATALINE,fdat)==NULL) break ;
				s = buf ;
//				printf("line=%s\n",s);
				for (i=0; i<CF_NCHAN; i++) {
					channel[i][j] = atoi(strtok_r((i==0?s:NULL)," \t\n\r",&s));
//					printf("ch_%s[%d] = %d ",
//						sc_apername(CF_NCHAN,i),j,channel[i][j]) ;
				}
//				printf("\n");
			}
			if (j<CF_BASENUMBER) {
				/* Too little data read: */
				printf("File END\n");
				break ;
			}
		}
#ifndef SPEEDTEST
#ifdef SC_LOGIDX
		printf("\nCompute scintillation log-indices:\n");
#else
		printf("\nCompute scintillation indices:\n");
#endif
#endif
		if ((err = sc_compute(CF_BASENUMBER,ncorr, 0))<0) {
		fprintf(stderr,sc_error(err)); return err; }

#ifndef SPEEDTEST
		printf("%4s  %8s  %8s  %8s  %8s  %8s\n",
			"Chan","sc_ind","sc_desi","mean","cmean","sig/cov");
		for (j=0; j<NAPCOMB(CF_NCHAN); j++) {
			printf("%2s  %8.1E  ",sc_apername(CF_NCHAN,j),
				sc_getidx(meas,SC_DSI)[j]);
			if (j<CF_NCHAN)
				printf("%8.1E  ", sc_getidx(meas,SC_DESI)[j]) ;
			else
				printf("%8s  ", "-") ;
			if (j<CF_NCHAN) 
				printf("%8.1E  %8.1E  %8.1E\n", 
					sc_getmean(0)[j], sc_getmean(1)[j],	sc_getsig(0)[j]) ;
			else
				printf("%8s  %8s  %8.1E\n","-","-", sc_getsig(0)[j]);
		}
#endif
		if (stattest) {
			printf("Stat test results:\n") ;
			printf("%4s  %8s\n", "Chan","sc_ind");
			for (j=0; j<CF_NCHAN; j++) {
				if ((err=sc_stattest(channel[j],CF_BASENUMBER,&(stind[j])))<0) {
					fprintf(stderr,"Stat test error %s\n",sc_error(err)); 
					return err; 
				}
				printf("%2s  %8.1E  ",sc_apername(CF_NCHAN,j), stind[j]);
			}
			printf("\n");
		}
		sprintf(stamp,"# M %8.5f",utc);
		if ((err = sc_writeidx(fidx,SC_DSI,stamp))<0) {
			fprintf(stderr,sc_error(err)); 
			return err; 
		}
		if ((err = sc_writemom(fmom,SC_MOM,stamp))<0) {
			fprintf(stderr,sc_error(err)); 
			return err; 
		}
		utc += cf_microexp * CF_BASENUMBER / 3600 ;
	}

	/* Average indices: */
	printf("\nAverage scintillation indices\n");
	if ((err = sc_average(0,SC_MAXLAG))<0) { 
		fprintf(stderr,sc_error(err)); 
		return err; 
	}
	printf("Ok. Here they are:\n");
	printf("Chan\tSI       e_SI      DESI      e_DESI\n");
	for (i=0; i<NAPCOMB(CF_NCHAN); i++) {
		printf("%2s\t%8.1E %8.1E ",sc_apername(CF_NCHAN,i),
				sc_getavgidx(SC_DSI,0)[i], sqrt(sc_geter2idx(SC_DSI,0)[i]));
		if (i<CF_NCHAN)
			printf("%8.1E %8.1E\n", sc_getavgidx(SC_DESI,0)[i], 
					sqrt(sc_geter2idx(SC_DESI,0)[i]));
		else
			printf("\n");
	}
	if ((err = sc_writeavgidx(favg,SC_DSI,0,"# I"))<0) {
		fprintf(stderr,sc_error(err)); 
		return err;
	}
	if ((err = sc_writeavgidx(favg,SC_AFLUX,0,"# F"))<0) {
		fprintf(stderr,sc_error(err)); 
		return err;
	}

	if (fdat!=NULL) {
		fdat = freopen(av[1],"rt",fdat);
	}
	/* Leave indices */
	sc_again() ;
	printf("Release module arrays:\n") ;
	sc_done() ;
	fclose(fidx) ;
	fclose(fmom) ;
	fclose(favg) ;
	printf("===================== %d ===================\n",test);

} /* of test-cycle */
if (fdat) fclose(fdat);
printf("Ok.\n");

return 0 ;
} /* end of main() */
#endif

/*---------------------------------------------------------------------*/
/*----- Static data storages accessible by functions sc_get...   ------*/
/*----- they are allocated by sc_init() and released be sc_done() -----*/
/*---- We use these global arrays to avoid their multiple -------------*/
/*---- allo/deallocation with all the times the same dimentions -------*/

static double ** sc_dsidx = NULL ;	/**< DSI matrix[Nmeas][Nidx] */
static double ** sc_desidx = NULL ;	/**< DESI matrix[Nmeas][Ndesi] */
static double * sc_adsi = NULL ;	/**< average DSI vector[Nidx*2] */
static double * sc_e2dsi = NULL ;	/**< error2 of avg. DSI vector[Nidx*2] */
static double * sc_adesi = NULL ;	/**< average DESI vector[Nchan*2] */
static double * sc_e2desi = NULL ;	/**< error2 of avg DESI vector[Nchan*2]*/
static double * sc_aflux = NULL ;	/**< average flux vector[Nchannel] */
static double * sc_e2flux = NULL ;	/**< error2 of avg.flux vector[Nchannel] */
static double * sc_mean = NULL;  	/**< mean counts vector[Nchan] */
static double ** sc_cmean = NULL ;	/**< corrected means matrix[Nchan][Nmeas] */
static double * sc_sig = NULL ;  	/**< count disp.and covar. vector[Nidx] */
static double * sc_r1 = NULL ;  	/**< lag-1 covar.vector[Nidx] */
static double * sc_r2 = NULL ;		/**< lag-2 covar.vector[Nchan] */
static sc_chan_t * sc_channel = NULL ;	/**< channel structure vector[Nchan] */

#ifdef SC_LOGIDX
static double * sc_logmean = NULL ;	/**< Mean ln(count) vector[Nchan] */
static double ** sc_logi = NULL ;	/**< ln(count) matrix[Nchan][Ndata] */
#endif

static signed char * sc_isgen = NULL ;   	/**< Shifted-mode flag vector[Nmeas] */

/* Memory is allocated in sc_init() and freed in sc_done() according to: */
static int sc_maxnmeas = 0 ;/**< Maximal number of measurements to store Nmeas*/
static int sc_maxndata = 0 ;/**< Maximal number of counts in buffers */
static int sc_nchan = 0 ;	/**< Current channels number Nchan */
static int sc_nindex = 0 ;	/**< Respective current Nidx as NAPCOMB(sc_nchan) */
/* actual number "nmeas" of measurements made is got from sc_curmeas; 
  actual number of counts put in buffer is given directly to sc_compute() */
static int sc_curmeas = -1 ;/**< Current measurement number */

/*-------------------------- sc_getcurmeas ----------------------------------*/

int sc_getcurmeas() {
	return sc_curmeas ;
}

/*-------------------------- sc_getidx ---------------------------------------*/

double * sc_getidx(int meas, sc_what what) {
	/* memory protection: */
	if ((meas>sc_maxnmeas)	|| (meas<0)) return NULL ;
	
	switch(what) {
		case SC_DSI:	return sc_dsidx[meas] ;
		case SC_DESI:	return sc_desidx[meas] ;
		default:    	return NULL ;
	}
}

/*-------------------------- sc_getavgidx ---------------------------------------*/

double * sc_getavgidx(sc_what what, signed char isgen) {

	switch(what) {
		case SC_DSI:	return sc_adsi  + isgen*sc_nindex ;
		case SC_DESI:	return sc_adesi + isgen*sc_nchan ;
		default:     	return NULL ;
	}
}

/*-------------------------- sc_geter2idx ---------------------------------------*/

double * sc_geter2idx(sc_what what, signed char isgen) {

	switch(what) {
		case SC_DSI:	return sc_e2dsi  + isgen*sc_nindex ;
		case SC_DESI:	return sc_e2desi + isgen*sc_nchan ;
		default:     	return NULL ;
	}
}

/*-------------------------- sc_getmean --------------------------------------*/

double * sc_getmean(int iscor) {	
	if (iscor) 
		return sc_cmean[sc_curmeas-1] ;
	else
		return sc_mean ;
}

/*-------------------------- sc_getavgflux -----------------------------------*/

double * sc_getavgflux(signed char isgen) {	
	return sc_aflux + isgen*sc_nchan ;
}

/*-------------------------- sc_getsig --------------------------------------*/

double * sc_getsig(int lag) {
	switch (lag) {
		case 0: return sc_sig ;
		case 1:	return sc_r1 ;
		case 2:	return sc_r2 ;
		default: return NULL ;		
	}
}

/*------------------------- sc_apername --------------------------------------*/
char * sc_apername(int nchan, int index) {
	register int i, j, n ;
	static char name[3] ;
	if ((index<0) || (nchan<0)) return NULL ;
	
	if (index<nchan) {
		/* Normal index: */
		name[1] = '\0';
		name[0] = (char)((int)'A' + index);
		return name ;
	} else {
		/* Differential index: */
		n = nchan;
		for (i=0;i<nchan-1;i++)
			for (j=i+1;j<nchan;j++)
				if ((n++)==index) {
					name[0] = (char)((int)'A' + i);
					name[1] = (char)((int)'A' + j);
					name[2] = '\0';
					return name ;
				}
	}
	/* If we drop here, then index > (number of apert. and their comb.) */
	return NULL;
}

/*-------------------------- count_matrix --------------------------------*/
/* Template utilities for matrix of the count_t type (de)allocation. Taken from
Numerical Recipes. It is not possible to use any of NR's ...matrix() since the 
particular type of count_t is not known */
//#define SC_MEMCHECK
count_t ** count_matrix(int nrl,int nrh,int ncl,int nch) {
	int i;
	count_t ** m;
 
	m=(count_t **) malloc((unsigned) (nrh-nrl+1)*sizeof(count_t*));
	if (!m) nrerror("allocation failure 1 in count_matrix()");
	m -= nrl;
 
	for(i=nrl;i<=nrh;i++) {
		m[i]=(count_t *) malloc((unsigned) (nch-ncl+1)*sizeof(count_t));
		if (!m[i]) nrerror("allocation failure 2 in count_matrix()");
		m[i] -= ncl;
	}
#ifdef SC_MEMCHECK
	printf("---------- allocate %x cnt_matrix[%d,%d][%d,%d]\n",(int)m,
		nrl,nrh,ncl,nch);
#endif
	return m;
}

/*---------------------- free_count_matrix ------------------------------*/

void free_count_matrix(count_t ** m, int nrl, int nrh, int ncl, int nch) {
	int i;

#ifdef SC_MEMCHECK
	printf("-------- DEallocate %x count_matrix[%d,%d][%d,%d]\n",(int)m,
		nrl,nrh,ncl,nch);
#endif 
	for(i=nrh;i>=nrl;i--) free((char*) (m[i]+ncl));
	free((char*) (m+nrl));
}

/*---------------------------- sc_init ----------------------------------*/

int sc_init(int ndata, int nmeas, int nchan) {
	int nindex, i ;

	if ((nmeas<=0) || (nchan<=0)) return SC_ERPAR ;

	/* If channel number is changed - redo all: */
	if (nchan>sc_nchan) {
		/* Free all memory if allocated: */
		if (sc_nchan>0)
			sc_done() ;
		
		/* Allocate vectors related to the number of channels only: */
		nindex = NAPCOMB(nchan) ;
		if (((sc_mean=dvector(0,nchan-1))==NULL) ||
#ifdef SC_LOGIDX
			((sc_logmean=dvector(0,nchan-1))==NULL) ||
#endif
			((sc_sig=dvector(0,nindex-1))==NULL) ||
			((sc_adsi=dvector(0,2*nindex-1))==NULL) ||
			((sc_e2dsi=dvector(0,2*nindex-1))==NULL) ||
			((sc_adesi=dvector(0,2*nchan-1))==NULL) ||
			((sc_e2desi=dvector(0,2*nchan-1))==NULL) ||
			((sc_aflux=dvector(0,2*nchan-1))==NULL) ||
			((sc_e2flux=dvector(0,2*nchan-1))==NULL) ||
			((sc_r1=dvector(0,nindex-1))==NULL) ||
			((sc_r2=dvector(0,nchan-1))==NULL) ||
				/* Channel structures: */
			((sc_channel = (sc_chan_t*)
				malloc((unsigned)nchan*sizeof(sc_chan_t)))==NULL))
				return SC_ERMEM ;
		sc_nchan = nchan ;
		sc_nindex = nindex ;
	} else {
		if (nchan!=sc_nchan) {
			sc_nchan = nchan ;
			sc_nindex = NAPCOMB(nchan) ;
		}
	}

	/* Data number in buffers changed: */
#ifdef SC_LOGIDX
	/* logarithms of counts storage: */
	if (sc_maxndata<ndata) {
		if (sc_maxndata)
			free_dmatrix(sc_logi,0,sc_nchan-1,0,sc_maxndata-1) ;
		if ((sc_logi=dmatrix(0,sc_nchan-1,0,ndata-1))==NULL) return SC_ERMEM ;
		sc_maxndata = ndata ;
	}
#endif		

	/* Maximal number of measurements changed: */
	if (sc_maxnmeas<nmeas) {
		/* Generalized mode flags: */
		if (sc_maxnmeas) {
			free(sc_isgen) ;
			free_dmatrix(sc_dsidx,0,sc_maxnmeas-1,0,sc_nindex-1) ;
			free_dmatrix(sc_desidx,0,sc_maxnmeas-1,0,sc_nchan-1) ;
			free_dmatrix(sc_cmean,0,sc_maxnmeas-1,0,sc_nchan-1) ;
		}
		
		if ((sc_isgen=(signed char *)malloc(nmeas*sizeof(signed char)))
			==NULL) return SC_ERMEM ;
		/* Index storages: */
		if (((sc_dsidx=dmatrix(0,nmeas-1,0,sc_nindex-1))==NULL) ||
			((sc_desidx=dmatrix(0,nmeas-1,0,sc_nchan-1))==NULL) ||
			((sc_cmean=dmatrix(0,nmeas-1,0,sc_nchan-1))==NULL)) 
				return SC_ERMEM ;
		sc_maxnmeas  = nmeas ;
	} 

	/* Initialize measurements: */
	for (i=0; i<sc_maxnmeas; i++) 
		sc_isgen[i] = (signed char)SC_MODENO ;	/* No measurement so far */
	sc_curmeas = 0 ;
	return 0 ;
}

/*---------------------------- sc_done ----------------------------------*/

void sc_done() {
	/* Check pointers to bear a signature "unallocated" - i.e. NULL, by
	our convention: */
	if (sc_nchan>0) {
		free_dvector(sc_mean,0,sc_nchan-1); sc_mean = NULL ;
		free_dvector(sc_sig,0,sc_nindex-1); sc_sig = NULL ;
		free_dvector(sc_adsi,0,2*sc_nindex-1) ; sc_adsi = NULL ;
		free_dvector(sc_e2dsi,0,2*sc_nindex-1) ; sc_e2dsi = NULL ;
		free_dvector(sc_adesi,0,2*sc_nchan-1) ; sc_adesi = NULL ;
		free_dvector(sc_e2desi,0,2*sc_nchan-1) ; sc_e2desi = NULL ;
		free_dvector(sc_aflux,0,2*sc_nchan-1) ; sc_aflux = NULL ;
		free_dvector(sc_e2flux,0,2*sc_nchan-1) ; sc_e2flux = NULL ;
		free_dvector(sc_r1,0,sc_nindex-1); sc_r1 = NULL; 
		free_dvector(sc_r2,0,sc_nchan-1); sc_r2 = NULL; 
#ifdef SC_LOGIDX
		free_dvector(sc_logmean,0,sc_nchan-1); sc_logmean = NULL; 
#endif
	}
#ifdef SC_LOGIDX
	if (sc_logi!=NULL)
		free_dmatrix(sc_logi,0,sc_nchan-1,0,sc_maxndata-1);	sc_logi = NULL; 
#endif
		
	if (sc_cmean!=NULL) { free_dmatrix(sc_cmean,0,sc_maxnmeas-1,0,sc_nchan-1); 
		sc_cmean = NULL;}

	if (sc_dsidx!=NULL) { free_dmatrix(sc_dsidx,0,sc_maxnmeas-1,0,sc_nindex-1); 
		sc_dsidx = NULL ; }
	if (sc_desidx!=NULL) { free_dmatrix(sc_desidx,0,sc_maxnmeas-1,0,sc_nchan-1); 
		sc_desidx = NULL ; }
	if (sc_isgen!=NULL) { free(sc_isgen) ; sc_isgen = NULL ; }

	/* Channel structures: */
	if (sc_channel != NULL) { free((char*) sc_channel) ; sc_channel = NULL ; }

	sc_curmeas = -1 ;	/* Signals impossibility to store any index */
	sc_nchan = sc_nindex = 0 ;
	sc_maxnmeas = 0 ;
	sc_maxndata = 0 ;
	return ;
}

/*---------------------------- sc_again ---------------------------------*/

void sc_again() {
	if (sc_curmeas>0)
		sc_curmeas = 0 ;
	return ;
}

/*---------------------------- sc_chan_init -----------------------------*/

int sc_chan_init(int chan, count_t * data, double nonpois, double bkgr, 
	double deadtime) {
	if ((chan<0) || (chan>=sc_nchan)) return SC_ERPAR ;
	if (sc_channel==NULL) return SC_ERNUL ;
	
	if (data!=NULL) sc_channel[chan].data = data;
	if (nonpois!=0) sc_channel[chan].nonpois = nonpois;
	if (bkgr!=0) sc_channel[chan].bkgr = bkgr;
	if (deadtime!=0) sc_channel[chan].deadtime = deadtime;
	return 0 ;
}

/*--------------------------- sc_writeidx --------------------------------*/

int sc_writeidx(FILE * f, sc_what what, const char * stamp) {

	int t,i ;

	switch(what) {
		case SC_HEAD:
			/* File needs the header only. Create header: */
			/* # <header> */
			if (stamp!=NULL) {
			if (!fprintf(f,"%s",stamp)) 
				/* Failed write attempt: */
				return SC_ERFIO ;
			}
			/* Plain indices header: */
			for (i=0;i<sc_nchan;i++) 
				fprintf(f,"%s SI_%s  ",SC_IDXDLM, sc_apername(sc_nchan,i));
			for (i=sc_nchan;i<sc_nindex;i++) 
				fprintf(f,"%s DSI_%s",SC_IDXDLM,sc_apername(sc_nchan,i));
		
			/* DESI header: */
			for (i=0;i<sc_nchan;i++) 
				fprintf(f,"%sDESI%d_%s",SC_IDXDLM, SC_DESIBIN, sc_apername(sc_nchan,i));
			break ;
		case SC_DSI :
		case SC_DESI :
			/* Check the first attempt to write results in something: */
			t = sc_curmeas-1 ;	/* Row number to take in index storages */
			if (stamp!=NULL) 
				if (fprintf(f,"%s",stamp)==0) return SC_ERFIO;
			for (i=0;i<sc_nindex;i++) 
				fprintf(f,SC_IDXFMT,SC_IDXDLM,sc_dsidx[t][i]);
			for (i=0;i<sc_nchan;i++) 
				fprintf(f,SC_IDXFMT,SC_IDXDLM,sc_desidx[t][i]);
			break ;
		default: 
			return SC_ERPAR ;
	}
	fprintf(f,"\n");
	return 0 ;
}

/*--------------------------- sc_writeavgidx ------------------------------*/

int sc_writeavgidx(FILE * f, sc_what what, signed char isgen, const char * stamp){
	int i ;

	/* Check the first attempt to write results in something: */
	if (stamp!=NULL) 
		if (fprintf(f,"%s",stamp)==0) return SC_ERFIO;

	switch(what) {
		case SC_AFLUX:
			/* Write the average fluxes: */
			for (i=0;i<sc_nchan;i++) {
				fprintf(f,SC_FLXFMT,SC_FLXDLM,sc_aflux[i+isgen*sc_nchan]);
				fprintf(f,SC_FLXFMT,SC_FLXDLM,sqrt(sc_e2flux[i+isgen*sc_nchan]));
			}
			break ;
		case SC_DSI:
		case SC_DESI:
			/* Write the average indices: */
			for (i=0;i<sc_nindex;i++) {
				fprintf(f,SC_IDXFMT,SC_IDXDLM,sc_getavgidx(SC_DSI,isgen)[i]);
				fprintf(f,SC_IDXFMT,SC_IDXDLM,sqrt(sc_geter2idx(SC_DSI,isgen)[i]));
			}
			for (i=0;i<sc_nchan;i++) {
				fprintf(f,SC_IDXFMT,SC_IDXDLM,sc_getavgidx(SC_DESI,isgen)[i]);
				fprintf(f,SC_IDXFMT,SC_IDXDLM,sqrt(sc_geter2idx(SC_DESI,isgen)[i]));
			}
			break ;
		default:
			return SC_ERPAR ;
	}
	fprintf(f,"\n");	
	return 0 ;
}

/*--------------------------- sc_writemom --------------------------------*/

int sc_writemom(FILE * f, sc_what what, const char * stamp) {

	int i ;

	switch(what) {
		case SC_HEAD:
			/* File needs the header only. Create header: */
			/* # <header> */
			if (stamp!=NULL) {
			if (!fprintf(f,"%s",stamp)) 
				/* Failed write attempt: */
				return SC_ERFIO ;
			}
			if (!fprintf(f,"%s",stamp)) 
				/* Failed write attempt: */
				return SC_ERFIO ;
				
			/* Nchan means: */
			for (i=0;i<sc_nchan;i++) 
				fprintf(f,"%sMEAN_%s",SC_MOMDLM, sc_apername(sc_nchan,i));
		
			/* Lag-0 dispersions and covariations header: */
			for (i=0;i<sc_nchan;i++) 
				fprintf(f,"%s SIG_%s ",SC_MOMDLM, sc_apername(sc_nchan,i));
			for (i=sc_nchan;i<sc_nindex;i++) 
				fprintf(f,"%s COV_%s",SC_MOMDLM, sc_apername(sc_nchan,i));
		
			/* Lag-1 auto- and cross-covariations: */
			for (i=0;i<sc_nindex;i++) 
				fprintf(f,"%s COV1_%2s",SC_MOMDLM, sc_apername(sc_nchan,i));
			/* Lag-2 autocovariations: */
			for (i=0;i<sc_nchan;i++) 
				fprintf(f,"%sCOV2_%s",SC_MOMDLM, sc_apername(sc_nchan,i));
			break ;
		case SC_MOM:
			/* Check the first attempt to write results in something: */
			if (stamp!=NULL)
				if (fprintf(f,"%s",stamp)==0) return SC_ERFIO;
			for (i=0;i<sc_nchan;i++) 
				fprintf(f,SC_MOMFMT,SC_MOMDLM,sc_mean[i]);
	
			for (i=0;i<sc_nindex;i++) 
				fprintf(f,SC_MOMFMT,SC_MOMDLM,sc_sig[i]);
			for (i=0;i<sc_nindex;i++) 
				fprintf(f,SC_MOMFMT,SC_MOMDLM,sc_r1[i]);
			for (i=0;i<sc_nchan;i++) 
				fprintf(f,SC_MOMFMT,SC_MOMDLM,sc_r2[i]);
			break ;
		default:
			return SC_ERPAR ;
	}
	fprintf(f,"\n");	
	return 0 ;
}

#ifdef SC_CNTMOD
/*--------------------- count data modifiers: ---------------------------*/
/*--------------------------- sc_smooth --------------------------------*/
/**
	@brief	Smooth the data counts with sliding window
	@param	data	data array
	@param	ndata	its length
	@param	band	width of smoothing window
	@param	smdata	Resulting filtered data array
	
	Function smooths the data counts in sliding window assigning the averaged
	counts to near-center elements of an output array. On the left and right
	margins, the window is cut. If \em band is even, then window is displaced
	by 1/2 pixels to the lower index (e.g.: band==2 gives a window containing
	the current pixel and pixels to the "left").

	\note
	- smdata[] array must contain at least ndata elements;
	
	@see sc_compute()
*/
static int sc_smooth(count_t * data, int ndata, int band, 
	count_t * smdata) {
	int i, lband, hband, k ; 
	count_t x ;
	register int j ;
	
	/* Verify the pointers and parameters: */
	if ((data == NULL) || (smdata == NULL)) return SC_ERNUL ;
	if (band<=0) return SC_ERPAR ;
	lband = band/2 ;	
	hband = band - 1 - lband ;
	/* left margin: */
	for (i=0; i<lband; i++) {
		x = 0 ;	k=0; for (j=0; j<=i+hband; j++) {k++; x += data[j] ;} 
		smdata[i] = x / k ;
	}
	/* middle of set: no marginal effects */
	for (i=lband; i<ndata-hband; i++) {
		x = 0 ;	for (j=i-lband; j<=i+hband; j++) x += data[j] ;
		smdata[i] = x / band ;
	}
	/* right margin: */
	for (i=ndata-hband; i<ndata; i++) {
		x = 0 ;	k=0; for (j=i-lband; j<ndata; j++) {k++; x += data[j] ;} 
		smdata[i] = x / k ;
	}
	return 0 ;
}

/*--------------------------- sc_bin --------------------------------*/
/**
	@brief	Bin the data array 
	@param	data	data array
	@param	ndata	its length
	@param	bin	binning factor (>0)
	@param	nbin	Resulting number of bins - length of output bin_data array
	@param	bin_data	Resulting binned data array
	@return	0 if Ok, <0 if error
	
	Procedure bins the data array producing one output value per each \em bin
	values of input arrays. Excess elements are ignored (e.g., 1000 input
	elements with 3-binning produce 333 output elements: 1 (last) element is
	thrown away). Number of elements is returned in nbin.
	
	\note
	- bin_data[] array must contain at least ndata elements.
	- providing bin=1 results in a plain copying of the input array into output
	
	@see sc_compute()
*/
static int sc_bin(count_t * data, int ndata, int bin, 
	int * nbin, count_t * bin_data) {
	int i ;
	count_t x ;
	register int j ;
	
	/* Verify the pointers and parameters: */
	if ((data == NULL) || (bin_data == NULL)) return SC_ERNUL ;
	if (bin<1) return SC_ERPAR ;

	*nbin = 0 ;
	for (i=0; i<ndata-bin+1; i+=bin) {
		x = 0;
		for (j=i; j<i+bin; j++) x += data[j] ;
		bin_data[*nbin] = x ;
		(*nbin)++ ;
	}
	return 0 ;
}
#endif

/*--------------------------- sc_indices --------------------------------*/

int sc_indices(sc_chan_t * sets, int nsets, int ndata, int * ncorr,
	double * scind, double * desi) {

#define LAG1 	1
#define LAG2	2	
/* Synonim of sc_sig here is sc_r0: */
#define sc_r0	sc_sig
	int ch, ch2, i, idx, ncomb ;
	double dx0=0, dx1, dx2, dy0, dy1 ;

	/* Verify the pointers and parameters: */
	if ((sets == NULL) || (scind == NULL)) return SC_ERNUL ;
	/* bin-factor cannot be neither non-positive nor too long: */
	if ((nsets<1) || (nsets>sc_nchan)) return SC_ERPAR ;
	/* Attention! last checking ensures that there will be no overfull in 
	double-type arrays of means, dispersions and covariations in channels */

	if (ndata<=LAG2) return SC_ERNOD ;
	
	/* Compute "raw" means */
	for (ch=0; ch<nsets; ch++) {
		dx1 = 0 ;
#ifdef SC_LOGIDX
		dx2 = 0 ;
		for (i=0; i<ndata; i++) {
			if (sets[ch].data[i]>0) {
				dx0 = log((double)sets[ch].data[i]) ;
				sc_logi[ch][i] = dx0 ;
				dx2 += dx0 ;
			} else {
				sc_logi[ch][i] = 0.0 ;
			}
			dx1 += sets[ch].data[i] ;
		}
		sc_logmean[ch] = dx2 / ndata ;
#else
		for (i=0; i<ndata; i++) dx1 += sets[ch].data[i] ;
#endif
		sc_mean[ch] = dx1 / ndata ;
//		printf("<log[%d]>=%f\n",ch,sc_logmean[ch]);
	}
	
	/* A cycle by channels: compute auto-correlations with lag up to 2
	(dispersions and auto-correlations with lags 1 and 2: */
	for (ch=0; ch<sc_nchan; ch++) {
		/* Compute "tails" which go out of cycle limits for r0 and r1: */
#ifdef SC_LOGIDX
		dx0 = sc_logi[ch][0] - sc_logmean[ch] ;
		dx1 = sc_logi[ch][1] - sc_logmean[ch] ;
#else
		dx0 = (double)sets[ch].data[0] - sc_mean[ch] ;
		dx1 = (double)sets[ch].data[1] - sc_mean[ch] ;
#endif
		sc_r0[ch] = dx0*dx0 + dx1*dx1 ;
		sc_r1[ch] = dx0*dx1 ;
		sc_r2[ch] = 0 ;
		for (i=LAG2; i<ndata; i++) {
#ifdef SC_LOGIDX
			dx2 = sc_logi[ch][i] - sc_logmean[ch] ;
#else
			dx2 = (double)sets[ch].data[i] - sc_mean[ch] ;
#endif
			sc_r0[ch] += dx2*dx2 ;
			sc_r1[ch] += dx2*dx1 ;
			sc_r2[ch] += dx2*dx0 ;
			dx0 = dx1 ;
			dx1 = dx2 ;
		}
		sc_r1[ch] /= (ndata-LAG1-ncorr[ch]) ;
		sc_r2[ch] /= (ndata-LAG2-ncorr[ch]) ;
		sc_r0[ch] /= (ndata-ncorr[ch]) ;
#ifdef SC_LOGIDX
		/* Denormalize by the squared mean of counts: */
		dx0 = (sc_mean[ch] * sc_mean[ch]) ;
		sc_r0[ch] *= dx0 ;
		sc_r1[ch] *= dx0 ;
		sc_r2[ch] *= dx0 ;
#endif
	} 

	/* A similar cycle by channels: compute auto-correlations with lag up to 1
	(dispersions and auto-correlations with lag 1 : */
	for (ch=sc_nchan; ch<nsets; ch++) {
		/* Compute "tails" which go out of cycle limits for r0: */
	#ifdef SC_LOGIDX
		dx0 = sc_logi[ch][0] - sc_logmean[ch] ;
	#else
		dx0 = (double)sets[ch].data[0] - sc_mean[ch] ;
	#endif
		sc_r0[ch] = dx0*dx0 ;
		sc_r1[ch] = 0 ;
		for (i=LAG1; i<ndata; i++) {
	#ifdef SC_LOGIDX
			dx1 = sc_logi[ch][i] - sc_logmean[ch] ;
	#else
			dx1 = (double)sets[ch].data[i] - sc_mean[ch] ;
	#endif
			sc_r0[ch] += dx1*dx1 ;
			sc_r1[ch] += dx0*dx1 ;
			dx0 = dx1 ;
		}
		sc_r1[ch] /= (ndata-LAG1) ;
		sc_r0[ch] /= ndata ;
	#ifdef SC_LOGIDX
		/* Denormalize by the squared mean of counts: */
		dx0 = (sc_mean[ch] * sc_mean[ch]) ;
		sc_r0[ch] *= dx0 ;
		sc_r1[ch] *= dx0 ;
	#endif
	} 

	/* Two embedded cycles by channels:  compute covariations */
	/* seq. number of channels combinations (channel1!=channel2) */
	ncomb = nsets; 
	for (ch=0; ch<nsets-1; ch++) {
		for (ch2=ch+1; ch2<nsets; ch2++) {
#ifdef SC_LOGIDX
			dx0 = sc_logi[ch ][0] - sc_logmean[ch ] ;
			dy0 = sc_logi[ch2][0] - sc_logmean[ch2] ;
#else
			dx0 = (double)sets[ch ].data[0] - sc_mean[ch ] ;
			dy0 = (double)sets[ch2].data[0] - sc_mean[ch2] ;
#endif
			sc_r0[ncomb] = dx0*dy0 ;
			sc_r1[ncomb] = 0 ;
			for (i=LAG1; i<ndata; i++) {
#ifdef SC_LOGIDX			
				dx1 = sc_logi[ch ][i] - sc_logmean[ch ] ;
				dy1 = sc_logi[ch2][i] - sc_logmean[ch2] ;
#else
				dx1 = (double)sets[ch ].data[i] - sc_mean[ch ] ;
				dy1 = (double)sets[ch2].data[i] - sc_mean[ch2] ;
#endif
				sc_r0[ncomb] += dx1*dy1 ;
				sc_r1[ncomb] += dx1*dy0+dx0*dy1 ;
				dx0 = dx1 ;
				dy0 = dy1 ;
			}
			sc_r1[ncomb] /= (2*(ndata-LAG1)) ;
			sc_r0[ncomb] /= ndata ;
#ifdef SC_LOGIDX
			/* Denormalize by the squared mean of counts: */
			dx0 = (sc_mean[ch] * sc_mean[ch2]) ;
			sc_r0[ncomb] *= dx0 ;
			sc_r1[ncomb] *= dx0 ;
#endif
			ncomb ++;
		}
	}

	/* Non-linearity correction: */
	ncomb = nsets ;
	for (ch=0;ch<nsets;ch++) {
		for (ch2=ch;ch2<nsets; ch2++) {
			if (ch==ch2) {
				idx = ch ;
				/* Correct means: */ 
				sc_cmean[sc_curmeas][ch] = 
				sc_mean[ch] * (1.0 + sc_mean[ch] * sets[ch].deadtime) ;
				dx0 = 1.0 - 2.0 * sc_mean[ch] * sets[ch].deadtime ;
				dy0 = dx0 ;
			} else {
				dy0 = 1.0 - 2.0 * sc_mean[ch2] * sets[ch2].deadtime ;
				idx = ncomb ;
				ncomb++ ;
			}
			dx2 = dx0*dy0 ;
			sc_r0[idx] /= dx2 ;
			sc_r1[idx] /= dx2 ;
			if (idx<sc_nchan) sc_r2[idx] /= dx2 ;
		}
	}

	/* Normal scintillation indices corrected to zero exposure: */
	for (ch=0; ch<nsets; ch++) {
		dx0 = (sc_cmean[sc_curmeas][ch] - sets[ch].bkgr) ;
		scind[ch] = (
					1.5*sc_r0[ch] - 0.5*sc_r1[ch] - 1.5 * 
						sc_cmean[sc_curmeas][ch] * sets[ch].nonpois) / (dx0*dx0) ;
	}

	/* Differential scintillation indices given the known already
		normal indices */
	ncomb = nsets ;
	for (ch=0;ch<nsets-1;ch++) {
		for (ch2=ch+1;ch2<nsets; ch2++) {
			dx0 = (sc_cmean[sc_curmeas][ch] - sets[ch].bkgr) ;
			dy0 = (sc_cmean[sc_curmeas][ch2] - sets[ch2].bkgr) ;
			scind[ncomb] = scind[ch] + scind[ch2] - 2.0 *
				(1.5*sc_r0[ncomb]-0.5*sc_r1[ncomb]) 
				/ (dx0 * dy0) ;
			ncomb ++;
		}
	}
	/* DESI */
	if (desi!=NULL)
		for (ch=0; ch<sc_nchan; ch++) {
			dx0 = (sc_cmean[sc_curmeas][ch] - sets[ch].bkgr) ;
			dx0 *= dx0 ;
			desi[ch] = (scind[ch]*3 - sc_r1[ch]*4/dx0 + sc_r2[ch]/dx0)/ 4.5 ; 
		}
	return 0 ;
#undef LAG1
#undef LAG2
#undef sc_r0
}

/*--------------------------- sc_compute --------------------------------*/

int sc_compute(int ndata, int * ncorr, signed char isgen) {

	int err ;
	/* Prevent the overflow of index storages: */
	if (sc_curmeas>=sc_maxnmeas) return SC_EROFL ;
	/* Prevent the call when no buffers are allocated: */
	if (sc_curmeas<0) return SC_ERPAR ;

	if ((err = sc_indices(sc_channel, sc_nchan, ndata, ncorr, 
		sc_dsidx[sc_curmeas], sc_desidx[sc_curmeas]))<0)
		return err ;
	
	sc_isgen[sc_curmeas] = (signed char)(isgen==SC_MODEGEN) ;
	
	sc_curmeas++ ;
	return 0 ;
}

/*--------------------------- sc_avgmatrix --------------------------------*/

int sc_avgmatrix(double ** matrix, int ncol, int nrow, 
	signed char * rowselect, signed char select, 
	double * avg, double * err2, int lag) {
	int t, i, k, nmeas ;
	double * rho = NULL ;
	double * dif = NULL ;

	/* Verify the pointers and parameters: */
	if ((matrix == NULL) || (avg == NULL) || (err2 == NULL)) return SC_ERNUL ;
		
	/* Allocate the storage for elementary differences */
	if ((dif = dvector(0, nrow-1))==NULL) return SC_ERMEM ;
	if ((rho = dvector(0, lag))==NULL) return SC_ERMEM ;

	/* Cycle by columns of the storage matrix: */
	for (i=0; i<ncol; i++) {
		avg[i] = 0 ;
		/* Cycle by base-time measurements : Simple mean accumulation */
		nmeas = 0;
		if  (ncol==1) {
			/* Treate the vector-like matrix transponed: */
			for (t=0; t<nrow; t++) 
				if ((rowselect==NULL) || (rowselect[t]==select)) {
					dif[nmeas] = matrix[i][t] ;
					avg[i] += dif[nmeas] ;
					nmeas ++ ;
				}
		} else {
			for (t=0; t<nrow; t++) 
				if ((rowselect==NULL) || (rowselect[t]==select)) {
					dif[nmeas] = matrix[t][i] ;
					avg[i] += dif[nmeas] ;
					nmeas ++ ;
				}
		}
		if (nmeas)
			avg[i] /= nmeas ;
		else {
			free_dvector(dif,0,nrow-1) ;
			free_dvector(rho,0,lag) ;
			return SC_ERNOD ;
		}

		/* Correct too high maximal lag: */
		if (lag>nmeas-1) lag = nmeas - 1 ;
	
		/* Produce the differences themselves: */
		for (t=0; t<nmeas; t++) 
			dif[t] -= avg[i] ;

		/* Compute the moments with lag up to "lag" */
		for (k=0; k<=lag; k++) {
			rho[k] = 0 ;
			for (t=0; t<nmeas-k; t++)
				rho[k] += dif[t] * dif[t+k] ;
		}
//#define PRNCORREL				
#ifdef PRNCORREL
if (lag>0) {
	printf("# corr.coef.for %d-th index (lag=1..%d):\n",i+1,lag);
	for (k=1; k<=lag; k++) printf("%6.4f ",rho[k]/rho[0]);
	printf("\n");
}
#endif

		/* Error of mean: Korn and Korn, section 19.8 "Statistics and
		measurement of the random process": "means by finit time span" */
		err2[i] = rho[0] / (nmeas-1) ;
		for (k=1; k<=lag; k++)
			err2[i] += 2*(1.0-(double)k/nmeas)*rho[k]/(nmeas-k) ;
		err2[i] /= nmeas ;
		
	}
	free_dvector(dif,0,nrow-1) ;
	free_dvector(rho,0,lag) ;
	return 0;
}

/*--------------------------- sc_average --------------------------------*/

int sc_average(signed char inclgen, int lag) {
	int err ;
	signed char isgen ;
#define NSETS 3

	for (isgen=0; isgen<=(inclgen>0); isgen++) {
	
		/* Average indices normal and differential aperture: */
		if ((err=sc_avgmatrix(sc_dsidx, sc_nindex, sc_getcurmeas(),
				sc_isgen, isgen, sc_adsi+isgen*sc_nindex, sc_e2dsi, lag))<0) 
				return err ;

		/* Average DESI: */
		if ((err=sc_avgmatrix(sc_desidx, sc_nchan, sc_getcurmeas(),
				sc_isgen, isgen, sc_adesi+isgen*sc_nchan, sc_e2desi, lag))<0) 
				return err ;
				
		/* Average corrected mean fluxes: */
		if ((err=sc_avgmatrix(sc_cmean,sc_nchan,sc_getcurmeas(),
				sc_isgen,isgen,sc_aflux+isgen*sc_nchan,sc_e2flux,lag))<0) 
				return err ;

		}
	return 0 ;
#undef NSETS
}

/*------------------------------ sc_calcstat ---------------------------------*/

int sc_calcstat(count_t * data, int n, int k, double * avg, double * sig, 
	int isfilt) {
	int i ;
	double d, avgdata, avgnew, sigdata ;

	if ((data==NULL) || (n<=0) || (k<1)) return SC_ERNUL ;

	/* Sum the fluxes in the current array of counts: */
	avgdata = 0 ;
	for (i=0; i<n; i++) 
		avgdata += (double) (data[i]) ;

	/* Average flux: */
	avgdata /= n ;
	
	/* Dispersion of the flux: [count^2]*/
	sigdata = 0 ;
	for (i=0; i<n; i++) {
		d = (double) (data[i]) - avgdata ;
		sigdata += d*d ;
	}
	if (n>2) sigdata /= (n-1) ;
	
	if (k==1) {
		/* First call to the function */
		*avg = avgdata ;
		*sig = sigdata ;
	} else {
		/* Not first call to the function with these avg and sig */
		(*sig) = ( (k-1) * (*sig) + sigdata) / k ;
		avgnew = ( (k-1) * (*avg) + avgdata) / k ;
		if (!isfilt) {
			d = (*avg) - avgnew ;
			(*sig) += (k-1) * d*d ;
		}
		(*avg) = avgnew ;
	}
	return 0 ;	
}

/*------------------------------ sc_stattest ---------------------------------*/

int sc_stattest(count_t * data, int ndata, double * scind) {
#define NTAKT	4	/* periodicity of the test signal */
#define TAKTPLUS 1	/* i.e. at each second takt the flux is increased */
#define TAKTMINUS 3	/* i.e. at each last takt the flux is decreased */
	int i, takt, nmid, nplus, nminus ;
	double fluxmid, fluxplus, fluxminus, fluxavg, dplus, dminus, dmid, sig, rho;

	if ((data==NULL) || (scind==NULL)) return SC_ERNUL ;
	if (ndata<=0) return SC_ERNOD ;
	
	fluxmid = fluxplus = fluxminus = 0 ;
	nmid = nplus = nminus = 0 ;
	/* Sum the fluxes in different takts: */
	for (i=0; i<ndata; i++) {
		takt = i % NTAKT ;
		switch (takt) {
			case TAKTPLUS:
				fluxplus += (double) (data[i]) ;
				nplus++ ;
				break ;
			case TAKTMINUS:
				fluxminus += (double) (data[i]) ;
				nminus++ ;
				break ;
			default:
				fluxmid += (double) (data[i]) ;
				nmid++ ;
		}
	}
	/* Average fluxes: */
	fluxplus /= nplus;
	fluxminus /= nminus ;
	fluxmid /= nmid ;
	fluxavg = (fluxplus*nplus + fluxminus*nminus + fluxmid*nmid) /
			(nplus + nminus + nmid) ;
	
	/* Deviations from the average level: */
	dplus = fluxplus - fluxavg ;
	dminus = fluxminus - fluxavg ;
	dmid = fluxmid - fluxavg ;
	
	/* Theoretic dispersion and lag=1,2 autocovariations: */
	sig  = (2*dmid*dmid + dplus*dplus + dminus*dminus) / 4 ;
	rho = dmid * (dplus + dminus) / 2 ;
//	rho2 = (dmid*dmid + dplus*dminus) / 2 ;
	
	/* Normal scintillation index: assume zero background */
	*scind = (1.5*sig - 0.5*rho) / fluxavg / fluxavg ;
	return 0 ;	
}

/*------------------------------- SC_ERROR --------------------------------*/

/* Error messages corresponding to SC_ER... codes: */
static char sc_ermsg[-SC_ERHI+1][60] = {
"NO ERROR",   /* 0: Ok */
"MEMORY allocation error",      		/* -1 */
"NULL pointer supplied", 				/* -2 */
"Allocation attempt on non-NULL array",	/* -3 */
"Non-sense supplied parameter",   		/* -4 */

"Bad file info or I/O error",    		/* -5 */
"Not expected count series", 			/* -6 */
"NO DATA supplied", 					/* -7 */
"WEIF: z-grids are different", 			/* -8 */
"WEIF: Spectrum is incompatible or bad",/* -9 */
"WEIF: Apertures number or geometry are different", /* -10 */
"ATMOS: Spectrum and Response curve do not coincide"/* -11 */
};

char * sc_error(int ercode) {

	if ((ercode<=SC_ERLO) && (ercode>=SC_ERHI)) {
		return sc_ermsg[-ercode];
	} else {
		return NULL;
	}
}


